#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Audio Captioning datasets for PyTorch."""

__author__ = "Étienne Labbé (Labbeti)"
__author_email__ = "labbeti.pub@gmail.com"
__docs__ = "Audio Captioning Datasets"
__docs_url__ = "https://aac-datasets.readthedocs.io/en/stable/"
__license__ = "MIT"
__maintainer__ = "Étienne Labbé (Labbeti)"
__name__ = "aac-datasets"
__status__ = "Development"
__version__ = "0.7.0"


from .datasets.audiocaps import AudioCaps
from .datasets.clotho import Clotho
from .datasets.interface import list_datasets_names, load_dataset
from .datasets.macs import MACS
from .datasets.wavcaps import WavCaps
from .utils.globals import (
    get_default_ffmpeg_path,
    get_default_root,
    get_default_ytdlp_path,
    set_default_ffmpeg_path,
    set_default_root,
    set_default_ytdlp_path,
)

__all__ = [
    "AudioCaps",
    "Clotho",
    "MACS",
    "WavCaps",
    "get_default_ffmpeg_path",
    "get_default_root",
    "get_default_ytdlp_path",
    "set_default_ffmpeg_path",
    "set_default_root",
    "set_default_ytdlp_path",
    "list_datasets_names",
    "load_dataset",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os.path as osp
import random
from argparse import ArgumentParser, Namespace
from typing import Dict, Iterable, Union

import pythonwrench as pw
import yaml

from aac_datasets.datasets.audiocaps import AudioCaps, AudioCapsCard
from aac_datasets.datasets.clotho import Clotho, ClothoCard
from aac_datasets.datasets.macs import MACS, MACSCard
from aac_datasets.datasets.wavcaps import WavCaps, WavCapsCard
from aac_datasets.utils.globals import get_default_root

DATASETS_NAMES = (AudioCapsCard.NAME, ClothoCard.NAME, MACSCard.NAME, WavCapsCard.NAME)

logger = logging.getLogger(__name__)


def check_directory(
    root: str,
    verbose: int = 0,
    datasets: Union[Iterable[str], str] = DATASETS_NAMES,
) -> Dict[str, Dict[str, int]]:
    """Check which datasets are installed in root.

    :param root: The directory to check.
    :param verbose: The verbose level. defaults to 0.
    :param datasets: The datasets to search in root directory. defaults to DATASETS_NAMES.
    :returns: A dictionary of datanames containing the length of each subset.
    """
    if isinstance(datasets, str):
        datasets = [datasets]
    else:
        datasets = list(datasets)

    data_infos = [
        (AudioCapsCard.NAME, AudioCaps),
        (ClothoCard.NAME, Clotho),
        (MACSCard.NAME, MACS),
        (WavCapsCard.NAME, WavCaps),
    ]
    data_infos = [
        (ds_name, class_) for ds_name, class_ in data_infos if ds_name in datasets
    ]

    if verbose >= 1:
        logger.info(f"Start searching datasets in root='{root}'.")

    all_found_dsets = {}
    for ds_name, ds_class in data_infos:
        if verbose >= 1:
            logger.info(f"Searching for {ds_name}...")

        found_dsets = {}
        for subset in ds_class.CARD.SUBSETS:
            try:
                ds = ds_class(root, subset, verbose=0)
                if len(ds) > 0:
                    # Try to load a random item
                    index = random.randint(0, len(ds) - 1)
                    ds[index]
                found_dsets[subset] = ds

            except RuntimeError:
                if verbose >= 2:
                    logger.info(f"Cannot find {ds_name}_{subset}.")

        if len(found_dsets) > 0:
            all_found_dsets[ds_name] = found_dsets

    if verbose >= 1:
        msg = f"Checking if audio files exists for {len(all_found_dsets)} datasets..."
        logger.info(msg)

    for ds_name, dsets in all_found_dsets.items():
        for subset, ds in dsets.items():
            fpaths = ds[:, "fpath"]
            is_valid = [osp.isfile(fpath) for fpath in fpaths]
            if not all(is_valid):
                logger.error(f"Cannot find all audio files for {ds_name}.{subset}.")
            else:
                logger.info(f"Dataset {ds_name}.{subset} is valid.")

    all_valids_lens = {
        ds_name: {subset: len(ds) for subset, ds in dsets.items()}
        for ds_name, dsets in all_found_dsets.items()
    }
    return all_valids_lens


def _get_main_check_args() -> Namespace:
    parser = ArgumentParser(description="Check datasets in specified directory.")

    parser.add_argument(
        "--root",
        type=str,
        default=get_default_root(),
        help="The path to the parent directory of the datasets.",
    )
    parser.add_argument(
        "--verbose",
        type=int,
        default=1,
        help="Verbose level of the script. 0 means silent mode, 1 is default mode and 2 add additional debugging outputs.",
    )
    parser.add_argument(
        "--datasets",
        type=str,
        nargs="+",
        default=DATASETS_NAMES,
        help="The datasets to check in root directory.",
    )

    args = parser.parse_args()
    return args


def _main_check() -> None:
    args = _get_main_check_args()
    pw.setup_logging_verbose("aac_datasets", args.verbose)

    if args.verbose >= 2:
        logger.debug(yaml.dump({"Arguments": args.__dict__}, sort_keys=False))

    valid_datasubsets = check_directory(args.root, args.verbose, args.datasets)

    if args.verbose >= 1:
        msg = f"Found {len(valid_datasubsets)}/{len(args.datasets)} dataset(s) in root='{args.root}':"
        print(msg)
        if len(valid_datasubsets) > 0:
            print(yaml.dump(valid_datasubsets, sort_keys=False))


if __name__ == "__main__":
    _main_check()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import logging
import os
import os.path as osp
import subprocess
import time
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, get_args

import pythonwrench as pw
import torchaudio
import tqdm
from torchwrench.hub.download import download_file
from typing_extensions import Literal

from aac_datasets.datasets.functional.common import DatasetCard, LinkInfo
from aac_datasets.utils.audioset_mapping import (
    download_audioset_mapping,
    load_audioset_mapping,
)
from aac_datasets.utils.globals import _get_ffmpeg_path, _get_root, _get_ytdlp_path

logger = logging.getLogger(__name__)

AudioCapsSubset = Literal["train", "val", "test", "train_fixed"]
AudioCapsVersion = Literal["v1", "v2"]


class AudioCapsCard(DatasetCard):
    ANNOTATIONS_CREATORS: Tuple[str, ...] = ("crowdsourced",)
    CAPTIONS_PER_AUDIO: Dict[AudioCapsSubset, int] = {
        "train": 1,
        "val": 5,
        "test": 5,
        "train_fixed": 1,
    }
    CITATION: str = r"""
    @inproceedings{kim_etal_2019_audiocaps,
        title        = {{A}udio{C}aps: Generating Captions for Audios in The Wild},
        author       = {Kim, Chris Dongjoo  and Kim, Byeongchang  and Lee, Hyunmin  and Kim, Gunhee},
        year         = 2019,
        month        = jun,
        booktitle    = {Proceedings of the 2019 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long and Short Papers)},
        publisher    = {Association for Computational Linguistics},
        address      = {Minneapolis, Minnesota},
        pages        = {119--132},
        doi          = {10.18653/v1/N19-1011},
        url          = {https://aclanthology.org/N19-1011},
    }
    """
    DEFAULT_SUBSET: AudioCapsSubset = "train"
    DEFAULT_VERSION: AudioCapsVersion = "v1"
    HOMEPAGE: str = "https://audiocaps.github.io/"
    LANGUAGE: Tuple[str, ...] = ("en",)
    LANGUAGE_DETAILS: Tuple[str, ...] = ("en-US",)
    NAME: str = "audiocaps"
    PRETTY_NAME: str = "AudioCaps"
    SIZE_CATEGORIES: Tuple[str, ...] = ("10K<n<100K",)
    SUBSETS: Tuple[AudioCapsSubset, ...] = get_args(AudioCapsSubset)
    TASK_CATEGORIES: Tuple[str, ...] = ("audio-to-text", "text-to-audio")
    VERSIONS: Tuple[AudioCapsVersion, ...] = get_args(AudioCapsVersion)


def load_audiocaps_dataset(
    # Common args
    root: Union[str, Path, None] = None,
    subset: AudioCapsSubset = AudioCapsCard.DEFAULT_SUBSET,
    verbose: int = 0,
    *,
    # AudioCaps-specific args
    audio_format: str = "flac",
    exclude_removed_audio: bool = True,
    sr: int = 32_000,
    with_tags: bool = False,
    version: AudioCapsVersion = AudioCapsCard.DEFAULT_VERSION,
) -> Tuple[Dict[str, List[Any]], Dict[int, str]]:
    """Load AudioCaps metadata.

    :param root: Dataset root directory.
        The data will be stored in the 'AUDIOCAPS' subdirectory.
        defaults to ".".
    :param subset: The subset of AudioCaps to use. Can be one of :attr:`~AudioCapsCard.SUBSETS`.
        defaults to "train".
    :param verbose: Verbose level.
        defaults to 0.

    :param audio_format: Audio format and extension name.
        defaults to "flac".
    :param exclude_removed_audio: If True, the dataset will exclude from the dataset the audio not downloaded from youtube (i.e. not present on disk).
        If False, invalid audios will return an empty tensor of shape (0,).
        defaults to True.
    :param sr: The sample rate used for audio files in the dataset (in Hz).
        Since original YouTube videos are recorded in various settings, this parameter allow to download allow audio files with a specific sample rate.
        defaults to 32000.
    :param with_tags: If True, load the tags from AudioSet dataset.
        Note: tags needs to be downloaded with download=True & with_tags=True before being used.
        defaults to False.
    :param version: The version of the dataset. Can be one of :attr:`~AudioCapsCard.VERSIONS`.
        defaults to 'v1'.

    :returns: A dictionnary of lists containing each metadata.
        Expected keys: "audiocaps_ids", "youtube_id", "start_time", "captions", "fname", "tags", "is_on_disk".
    """
    if subset in _AUDIOCAPS_OLD_SUBSETS_NAMES:
        new_subset = _AUDIOCAPS_OLD_SUBSETS_NAMES[subset]
        if verbose >= 0:
            msg = f"Deprecated subset name '{subset}', use '{new_subset}' instead."
            logger.warning(msg)
        subset = new_subset

    root = _get_root(root)
    audiocaps_root = _get_audiocaps_root(root, sr, version)
    audio_subset_dpath = _get_audio_subset_dpath(root, subset, sr, version)

    is_prepared = _is_prepared_audiocaps(
        root,
        subset,
        sr,
        audio_format,
        verbose=verbose,
        version=version,
    )
    if not is_prepared:
        msg = f"Cannot load data: audiocaps_{subset} is not prepared in data {root=}. Please use download=True in dataset constructor."
        raise RuntimeError(msg)

    version_links = _AUDIOCAPS_LINKS[version]
    if subset not in version_links:
        msg = f"Subset {subset} is not available for AudioCaps version {version}. (expected one of {tuple(version_links.keys())})"
        raise ValueError(msg)

    links = version_links[subset]

    captions_dpath = _get_captions_dpath(root, subset, sr, version)
    captions_fname = links["captions"]["fname"]
    captions_fpath = osp.join(captions_dpath, captions_fname)

    with open(captions_fpath, "r") as file:
        reader = csv.DictReader(file)
        captions_data = list(reader)

    if with_tags:
        class_labels_indices_fpath = osp.join(
            audiocaps_root, _AUDIOSET_LINKS["class_labels_indices"]["fname"]
        )
        unbal_tags_fpath = osp.join(
            audiocaps_root, _AUDIOSET_LINKS["unbalanced"]["fname"]
        )

        if not all(map(osp.isfile, (class_labels_indices_fpath, unbal_tags_fpath))):
            msg = (
                f"Cannot load tags without tags files '{osp.basename(class_labels_indices_fpath)}' and '{osp.basename(unbal_tags_fpath)}'."
                + "Please use download=True and with_tags=True in dataset constructor."
            )
            raise FileNotFoundError(msg)

        mid_to_index: Dict[str, int] = load_audioset_mapping(
            "mid",
            "index",
            offline=True,
            cache_path=audiocaps_root,
            verbose=verbose,
        )
        index_to_name: Dict[int, str] = load_audioset_mapping(
            "index",
            "display_name",
            offline=True,
            cache_path=audiocaps_root,
            verbose=verbose,
        )

        with open(unbal_tags_fpath, "r") as file:
            FIELDNAMES = ("YTID", "start_seconds", "end_seconds", "positive_labels")
            reader = csv.DictReader(
                file,
                FIELDNAMES,
                skipinitialspace=True,
                strict=True,
            )
            # Skip the comments
            for _ in range(3):
                next(reader)
            unbal_tags_data = list(reader)
    else:
        mid_to_index = {}
        index_to_name = {}
        unbal_tags_data = []

    # Build global mappings
    fnames_dic = dict.fromkeys(
        _AUDIO_FNAME_FORMAT.format(**line, audio_format=audio_format)
        for line in captions_data
    )
    audio_fnames_on_disk = dict.fromkeys(os.listdir(audio_subset_dpath))
    if exclude_removed_audio:
        fnames_lst = [fname for fname in fnames_dic if fname in audio_fnames_on_disk]
        is_on_disk_lst = [True for _ in range(len(fnames_lst))]
    else:
        fnames_lst = list(fnames_dic)
        is_on_disk_lst = [fname in audio_fnames_on_disk for fname in fnames_lst]

    dataset_size = len(fnames_lst)
    fname_to_idx = {fname: i for i, fname in enumerate(fnames_lst)}

    # Process each field into a single structure
    all_caps_dic: Dict[str, List[Any]] = {
        key: [None for _ in range(dataset_size)]
        for key in ("audiocaps_ids", "youtube_id", "start_time", "captions")
    }
    for line in tqdm.tqdm(
        captions_data,
        disable=verbose <= 0,
        desc=f"Loading AudioCaps ({subset}) captions...",
    ):
        # audiocap_id, youtube_id, start_time, caption
        audiocap_id = line["audiocap_id"]
        youtube_id = line["youtube_id"]
        start_time = line["start_time"]
        caption = line["caption"]

        fname = _AUDIO_FNAME_FORMAT.format(**line, audio_format=audio_format)
        if fname in fname_to_idx:
            index = fname_to_idx[fname]

            if all_caps_dic["start_time"][index] is None:
                all_caps_dic["start_time"][index] = start_time
                all_caps_dic["youtube_id"][index] = youtube_id
                all_caps_dic["audiocaps_ids"][index] = [audiocap_id]
                all_caps_dic["captions"][index] = [caption]
            else:
                # sanity check
                assert all_caps_dic["start_time"][index] == start_time
                assert all_caps_dic["youtube_id"][index] == youtube_id

                all_caps_dic["audiocaps_ids"][index].append(audiocap_id)
                all_caps_dic["captions"][index].append(caption)

    # Load tags from audioset data
    all_tags_lst = [[] for _ in range(dataset_size)]

    for line in tqdm.tqdm(
        unbal_tags_data,
        disable=verbose <= 0,
        desc="Loading AudioSet tags for AudioCaps...",
    ):
        # keys: YTID, start_seconds, end_seconds, positive_labels
        youtube_id = line["YTID"]
        # Note : In audioset, start_time is a string repr of a float value, audiocaps it is a string repr of an integer
        start_time = int(float(line["start_seconds"]))
        fname = _AUDIO_FNAME_FORMAT.format(
            youtube_id=youtube_id,
            start_time=start_time,
            audio_format=audio_format,
        )
        if fname in fname_to_idx:
            tags_mid = line["positive_labels"]
            tags_mid = tags_mid.split(",")
            tags_indexes = [mid_to_index[tag_mid] for tag_mid in tags_mid]

            index = fname_to_idx[fname]
            all_tags_lst[index] = tags_indexes

    raw_data = {
        "fname": fnames_lst,
        "tags": all_tags_lst,
        "is_on_disk": is_on_disk_lst,
    }
    raw_data.update(all_caps_dic)

    # Convert audiocaps_ids and start_time to ints
    raw_data["audiocaps_ids"] = [
        list(map(int, item)) for item in raw_data["audiocaps_ids"]
    ]
    raw_data["start_time"] = list(map(int, raw_data["start_time"]))

    if verbose >= 1:
        msg = f"{AudioCapsCard.PRETTY_NAME}(subset={subset}) has been loaded. {len(fnames_lst)=})"
        logger.info(msg)

    return raw_data, index_to_name


def download_audiocaps_dataset(
    # Common args
    root: Union[str, Path, None] = None,
    subset: AudioCapsSubset = AudioCapsCard.DEFAULT_SUBSET,
    force: bool = False,
    verbose: int = 0,
    verify_files: bool = False,
    *,
    # AudioCaps-specific args
    audio_duration: float = 10.0,
    audio_format: str = "flac",
    audio_n_channels: int = 1,
    download_audio: bool = True,
    ffmpeg_path: Union[str, Path, None] = None,
    max_workers: Optional[int] = 1,
    sr: int = 32_000,
    ytdlp_path: Union[str, Path, None] = None,
    with_tags: bool = False,
    version: AudioCapsVersion = AudioCapsCard.DEFAULT_VERSION,
    ytdlp_opts: Iterable[str] = (),
    num_dl_attempts: int = 2,
) -> None:
    """Prepare AudioCaps data (audio, labels, metadata).

    :param root: Dataset root directory.
        The data will be stored in the 'AUDIOCAPS' subdirectory.
        defaults to ".".
    :param subset: The subset of AudioCaps to use. Can be one of :attr:`~AudioCapsCard.SUBSETS`.
        defaults to "train".
    :param force: If True, force to re-download file even if they exists on disk.
        defaults to False.
    :param verbose: Verbose level.
        defaults to 0.
    :param verify_files: If True, check hash value when possible.
        defaults to True.

    :param audio_duration: Extracted duration for each audio file in seconds.
        defaults to 10.0.
    :param audio_format: Audio format and extension name.
        defaults to "flac".
    :param audio_n_channels: Number of channels extracted for each audio file.
        defaults to 1.
    :param download_audio: If True, download audio, metadata and labels files. Otherwise it will only donwload metadata and labels files.
        defaults to True.
    :param ffmpeg_path: Path to ffmpeg executable file.
        defaults to "ffmpeg".
    :param max_workers: Number of threads to download audio files in parallel.
        Do not use a value too high to avoid "Too Many Requests" error.
        The value None will use `min(32, os.cpu_count() + 4)` workers, which is the default of ThreadPoolExecutor.
        defaults to 1.
    :param sr: The sample rate used for audio files in the dataset (in Hz).
        Since original YouTube videos are recorded in various settings, this parameter allow to download allow audio files with a specific sample rate.
        defaults to 32000.
    :param with_tags: If True, download the tags from AudioSet dataset.
        defaults to False.
    :param ytdlp_path: Path to yt-dlp or ytdlp executable.
        defaults to "yt-dlp".
    :param version: The version of the dataset. Can be one of :attr:`~AudioCapsCard.VERSIONS`.
        defaults to 'v1'.
    :param ytdlp_opts: yt-dlp options.
        defaults to ().
    :param num_dl_attempts: Number of download attempts.
        defaults to 2.
    """

    root = _get_root(root)
    ytdlp_path = _get_ytdlp_path(ytdlp_path)
    ffmpeg_path = _get_ffmpeg_path(ffmpeg_path)

    if not osp.isdir(root):
        raise RuntimeError(f"Cannot find root directory '{root}'.")

    _check_subprog_help(ytdlp_path, "ytdlp")
    _check_subprog_help(ffmpeg_path, "ffmpeg")

    is_prepared = _is_prepared_audiocaps(
        root, subset, sr, audio_format=audio_format, verbose=-1, version=version
    )
    if is_prepared and not force:
        return None

    audiocaps_root = _get_audiocaps_root(root, sr, version)
    os.makedirs(audiocaps_root, exist_ok=True)
    if with_tags:
        _download_tags_files(root, sr, version, verbose)

    version_links = _AUDIOCAPS_LINKS[version]
    if subset not in version_links:
        msg = f"Subset {subset} is not available for AudioCaps version {version}. (expected one of {tuple(version_links.keys())})"
        raise ValueError(msg)
    links = version_links[subset]

    audio_subset_dpath = _get_audio_subset_dpath(root, subset, sr, version)
    os.makedirs(audio_subset_dpath, exist_ok=True)

    captions_dpath = _get_captions_dpath(root, subset, sr, version)
    os.makedirs(captions_dpath, exist_ok=True)

    captions_fname = links["captions"]["fname"]
    captions_fpath = osp.join(captions_dpath, captions_fname)

    if not osp.isfile(captions_fpath):
        url = links["captions"]["url"]
        if url is None:
            msg = f"AudioCaps subset '{subset}' cannot be automatically downloaded. (found {url=})"
            raise ValueError(msg)

        download_file(url, captions_fpath, verbose=verbose)

    if download_audio:
        _download_audio_files(
            subset=subset,
            verbose=verbose,
            verify_files=verify_files,
            captions_fpath=captions_fpath,
            audio_subset_dpath=audio_subset_dpath,
            audio_duration=audio_duration,
            audio_format=audio_format,
            audio_n_channels=audio_n_channels,
            ffmpeg_path=ffmpeg_path,
            max_workers=max_workers,
            sr=sr,
            ytdlp_path=ytdlp_path,
            ytdlp_opts=ytdlp_opts,
            num_dl_attempts=num_dl_attempts,
        )

    if verbose >= 2:
        msg = f"Dataset {AudioCapsCard.PRETTY_NAME} {subset=}) has been prepared."
        logger.debug(msg)


def download_audiocaps_datasets(
    # Common args
    root: Union[str, Path, None] = None,
    subsets: Union[
        AudioCapsSubset, Iterable[AudioCapsSubset]
    ] = AudioCapsCard.DEFAULT_SUBSET,
    force: bool = False,
    verbose: int = 0,
    verify_files: bool = False,
    *,
    # AudioCaps-specific args
    audio_duration: float = 10.0,
    audio_format: str = "flac",
    audio_n_channels: int = 1,
    download_audio: bool = True,
    ffmpeg_path: Union[str, Path, None] = None,
    max_workers: Optional[int] = 1,
    sr: int = 32_000,
    with_tags: bool = False,
    ytdlp_path: Union[str, Path, None] = None,
    version: AudioCapsVersion = AudioCapsCard.DEFAULT_VERSION,
    ytdlp_opts: Iterable[str] = (),
    num_dl_attempts: int = 2,
) -> None:
    """Function helper to download a list of subsets. See :func:`~aac_datasets.datasets.functional.audiocaps.download_audiocaps_dataset` for details."""
    if isinstance(subsets, str):
        subsets = [subsets]
    else:
        subsets = list(subsets)

    kwargs: Dict[str, Any] = dict(
        root=root,
        force=force,
        verbose=verbose,
        verify_files=verify_files,
        audio_duration=audio_duration,
        audio_format=audio_format,
        audio_n_channels=audio_n_channels,
        download_audio=download_audio,
        ffmpeg_path=ffmpeg_path,
        max_workers=max_workers,
        sr=sr,
        with_tags=with_tags,
        ytdlp_path=ytdlp_path,
        version=version,
        ytdlp_opts=ytdlp_opts,
        num_dl_attempts=num_dl_attempts,
    )
    for subset in subsets:
        download_audiocaps_dataset(
            subset=subset,
            **kwargs,
        )


def _download_audio_files(
    subset: AudioCapsSubset,
    verbose: int,
    verify_files: bool,
    *,
    captions_fpath: str,
    audio_subset_dpath: str,
    audio_duration: float,
    audio_format: str,
    audio_n_channels: int,
    ffmpeg_path: str,
    max_workers: Optional[int],
    sr: int,
    ytdlp_path: str,
    ytdlp_opts: Iterable[str],
    num_dl_attempts: int = 2,
) -> None:
    start = time.perf_counter()
    if verbose >= 1:
        logger.info(f"Start downloading audio files for AudioCaps {subset} split...")

    with open(captions_fpath, "r") as file:
        # Download audio files
        reader = csv.DictReader(file)
        # Keys: audiocap_id, youtube_id, start_time, caption
        captions_data = list(reader)

    def _cast_line(line: Dict[str, Any], audio_format: str) -> Dict[str, Any]:
        youtube_id = line["youtube_id"]
        start_time = line["start_time"]

        if not start_time.isdigit():
            msg = f"Start time '{start_time}' is not an integer (with {youtube_id=})."
            raise RuntimeError(msg)

        start_time = int(start_time)
        fname = _AUDIO_FNAME_FORMAT.format(
            youtube_id=youtube_id,
            start_time=start_time,
            audio_format=audio_format,
        )
        line = pw.union_dicts([line, {"start_time": start_time, "fname": fname}])
        return line

    captions_data = [_cast_line(line, audio_format) for line in captions_data]
    download_kwds = {
        line["fname"]: {k: line[k] for k in ("fname", "youtube_id", "start_time")}
        for line in captions_data
    }
    del captions_data

    present_audio_fnames = os.listdir(audio_subset_dpath)
    present_audio_fpaths = [
        osp.join(audio_subset_dpath, fname) for fname in present_audio_fnames
    ]
    present_audio_fpaths = dict.fromkeys(present_audio_fpaths)

    common_kwds: Dict[str, Any] = dict(
        audio_subset_dpath=audio_subset_dpath,
        verify_files=verify_files,
        present_audio_fpaths=present_audio_fpaths,
        audio_duration=audio_duration,
        sr=sr,
        audio_n_channels=audio_n_channels,
        ffmpeg_path=ffmpeg_path,
        ytdlp_path=ytdlp_path,
        verbose=verbose,
        ytdlp_opts=ytdlp_opts,
        num_dl_attempts=num_dl_attempts,
    )
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        if verbose >= 2:
            logger.debug(f"Using {executor._max_workers} workers.")

        submitted_dict = {
            fname: executor.submit(
                _download_from_youtube_and_verify,
                **kwds,
                **common_kwds,
            )
            for fname, kwds in download_kwds.items()
        }
        for i, (fname, submitted) in enumerate(
            tqdm.tqdm(submitted_dict.items(), disable=verbose < 1)
        ):
            file_exists, download_success, valid_file = submitted.result()

            if verbose < 2:
                continue

            if not file_exists:
                if not download_success:
                    msg_end = f"File '{fname}' cannot be downloaded. (maybe the source video has been removed?)"
                elif valid_file:
                    msg_end = f"File '{fname}' has been downloaded and verified."
                elif verify_files:
                    msg_end = f"File '{fname}' has been downloaded but it was not valid and has been removed."
                else:
                    msg_end = f"File '{fname}' has been downloaded."
            else:
                if valid_file:
                    msg_end = (
                        f"File '{fname}' is already downloaded and has been verified."
                    )
                elif verify_files:
                    msg_end = f"File '{fname}' is already downloaded but it was not valid and has been removed."
                else:
                    msg_end = f"File '{fname}' is already downloaded."

            logger.debug(f"[{i + 1:5d}/{len(download_kwds)}] {msg_end}")

    if verbose >= 1:
        duration_s = int(time.perf_counter() - start)
        msgs = (
            f"Download and preparation of AudioCaps for subset '{subset}' done in {duration_s}s.",
            f"- {len(download_kwds)} total samples.",
        )
        logger.info("\n".join(msgs))


def _download_tags_files(
    root: Union[str, Path, None],
    sr: int,
    version: AudioCapsVersion,
    verbose: int,
) -> None:
    root = _get_root(root)
    audiocaps_root = _get_audiocaps_root(root, sr, version)

    target = "unbalanced"
    infos = _AUDIOSET_LINKS[target]
    url = infos["url"]
    fname = infos["fname"]
    fpath = osp.join(audiocaps_root, fname)
    if not osp.isfile(fpath):
        if verbose >= 1:
            logger.info(f"Downloading file '{fname}'...")
        download_file(url, fpath, verbose=verbose)

    download_audioset_mapping(audiocaps_root, verbose=verbose)


def _get_audiocaps_root(root: str, sr: int, version: AudioCapsVersion) -> str:
    return osp.join(root, "AUDIOCAPS")


def _get_audio_subset_dpath(
    root: str,
    subset: AudioCapsSubset,
    sr: int,
    version: AudioCapsVersion,
) -> str:
    return osp.join(
        _get_audiocaps_root(root, sr, version),
        f"audio_{sr}Hz",
        _AUDIOCAPS_AUDIO_DNAMES[subset],
    )


def _get_captions_dpath(
    root: str,
    subset: AudioCapsSubset,
    sr: int,
    version: AudioCapsVersion,
) -> str:
    audiocaps_root = _get_audiocaps_root(root, sr, version)
    captions_fname = _AUDIOCAPS_LINKS[version][subset]["captions"]["fname"]
    captions_fpath = osp.join(audiocaps_root, captions_fname)
    # For backward compatibility only
    if version == "v1" and osp.isfile(captions_fpath):
        return audiocaps_root
    else:
        return osp.join(audiocaps_root, f"csv_files_{version}")


def _is_prepared_audiocaps(
    root: str,
    subset: AudioCapsSubset = AudioCapsCard.DEFAULT_SUBSET,
    sr: int = 32_000,
    audio_format: str = "flac",
    verbose: int = 0,
    version: AudioCapsVersion = AudioCapsCard.DEFAULT_VERSION,
) -> bool:
    version_links = _AUDIOCAPS_LINKS[version]
    if subset not in version_links:
        msg = f"Subset {subset} is not available for AudioCaps version {version}. (expected one of {tuple(version_links.keys())})"
        raise ValueError(msg)
    links = version_links[subset]

    captions_dpath = _get_captions_dpath(root, subset, sr, version)
    captions_fname = links["captions"]["fname"]
    captions_fpath = osp.join(captions_dpath, captions_fname)
    audio_subset_dpath = _get_audio_subset_dpath(root, subset, sr, version)

    msgs = []

    if not osp.isdir(audio_subset_dpath):
        msg = f"Cannot find directory '{audio_subset_dpath}'."
        msgs.append(msg)
    else:
        audio_fnames = os.listdir(audio_subset_dpath)
        audio_fnames = [fname for fname in audio_fnames if fname.endswith(audio_format)]
        if len(audio_fnames) == 0:
            msg = (
                f"Cannot find any audio {audio_format} file in '{audio_subset_dpath}'."
            )
            msgs.append(msg)

    if not osp.isfile(captions_fpath):
        msg = f"Cannot find file '{captions_fpath}'."
        msgs.append(msg)

    if verbose >= 0:
        for msg in msgs:
            logger.warning(msg)

    return len(msgs) == 0


def _download_from_youtube_and_verify(
    fname: str,
    youtube_id: str,
    start_time: int,
    audio_subset_dpath: str,
    verify_files: bool,
    present_audio_fpaths: Dict[str, None],
    audio_duration: float,
    sr: int,
    audio_n_channels: int,
    ffmpeg_path: str,
    ytdlp_path: str,
    verbose: int,
    ytdlp_opts: Iterable[str],
    num_dl_attempts: int,
) -> Tuple[bool, bool, bool]:
    fpath = osp.join(audio_subset_dpath, fname)

    file_exists = fpath in present_audio_fpaths
    download_success = False
    valid_file = False

    if not file_exists:
        download_success = False

        for _ in range(num_dl_attempts):
            download_success = _download_from_youtube(
                youtube_id=youtube_id,
                fpath_out=fpath,
                start_time=start_time,
                audio_duration=audio_duration,
                sr=sr,
                audio_n_channels=audio_n_channels,
                ffmpeg_path=ffmpeg_path,
                ytdlp_path=ytdlp_path,
                verbose=verbose,
                ytdlp_opts=ytdlp_opts,
            )
            if download_success:
                break

            sleep_duration = 0.5
            if sleep_duration > 0.0 and verbose >= 2:
                logger.debug(f"Download failed, retrying in {sleep_duration:.1f}s...")
            time.sleep(sleep_duration)

    if verify_files and (download_success or file_exists):
        valid_file = _is_valid_audio_file(
            fpath,
            min_n_frames=1,
            sr=sr,
            n_channels=audio_n_channels,
        )

    if verify_files and not valid_file and osp.isfile(fpath):
        os.remove(fpath)

    return file_exists, download_success, valid_file


def _download_from_youtube(
    youtube_id: str,
    fpath_out: str,
    start_time: int,
    audio_duration: float = 10.0,
    sr: int = 32_000,
    audio_n_channels: int = 1,
    audio_format: str = "flac",
    acodec: str = "flac",
    ytdlp_path: Union[str, Path, None] = None,
    ffmpeg_path: Union[str, Path, None] = None,
    verbose: int = 0,
    ytdlp_opts: Iterable[str] = (),
) -> bool:
    """Download audio from youtube with yt-dlp and ffmpeg."""
    ytdlp_path = _get_ytdlp_path(ytdlp_path)
    ffmpeg_path = _get_ffmpeg_path(ffmpeg_path)
    ytdlp_opts = list(ytdlp_opts)

    # Get audio download link with yt-dlp, without start time
    link = _get_youtube_link(youtube_id, None)
    get_url_command = [
        ytdlp_path,
        "--youtube-skip-dash-manifest",
        "-g",
        link,
    ] + ytdlp_opts
    try:
        output = subprocess.check_output(get_url_command)
    except (CalledProcessError, PermissionError) as err:
        if verbose >= 2:
            logger.debug(err)
        return False

    output = output.decode()
    lines = output.split("\n")
    if len(lines) < 2:
        return False
    _video_link, audio_link = lines[:2]

    # if yt-dlp only returns one link, it is a combined video audio
    if len(audio_link) == 0:
        if verbose >= 2:
            msg = f"Video with youtube_id={youtube_id} is a combined video audio only that cannot be downloaded."
            logger.debug(msg)
        # audio_link = video_link # this does not work, not sure why. probably requires changes to ffmpeg command
        return False

    # Download and extract audio from audio_link to fpath_out with ffmpeg
    extract_command = [
        ffmpeg_path,
        # Input
        "-i",
        audio_link,
        # Remove video
        "-vn",
        # Format (flac)
        "-f",
        audio_format,
        # Audio codec (flac)
        "-acodec",
        acodec,
        # Get only 10s of the clip after start_time
        "-ss",
        str(start_time),
        "-t",
        str(audio_duration),
        # Resample to a specific rate (default to 32 kHz)
        "-ar",
        str(sr),
        # Compute mean of 2 channels
        "-ac",
        str(audio_n_channels),
        fpath_out,
    ]
    try:
        if verbose < 3:
            stdout = subprocess.DEVNULL
            stderr = subprocess.DEVNULL
        else:
            stdout = None
            stderr = None
        exitcode = subprocess.check_call(extract_command, stdout=stdout, stderr=stderr)
        return exitcode == 0

    except (CalledProcessError, PermissionError) as err:
        if verbose >= 2:
            logger.debug(err)
        return False


def _check_subprog_help(
    path: str,
    name: str,
    stdout: Any = subprocess.DEVNULL,
    stderr: Any = subprocess.DEVNULL,
) -> None:
    try:
        subprocess.check_call(
            [path, "--help"],
            stdout=stdout,
            stderr=stderr,
        )
    except (CalledProcessError, PermissionError, FileNotFoundError) as err:
        logger.error(f"Invalid {name} path '{path}'. ({err})")
        raise err


def _is_valid_audio_file(
    fpath: str,
    *,
    min_n_frames: Optional[int] = None,
    max_n_frames: Optional[int] = None,
    sr: Optional[int] = None,
    n_channels: Optional[int] = None,
) -> bool:
    try:
        metadata = torchaudio.info(fpath)  # type: ignore
    except RuntimeError:
        msg = f"Found file '{fpath}' already downloaded but it is invalid (cannot load metadata)."
        logger.error(msg)
        return False

    msgs = []
    if min_n_frames is not None and metadata.num_frames < min_n_frames:
        msg = f"Found file '{fpath}' already downloaded but it is invalid (audio is shorter than {min_n_frames=} samples)."
        msgs.append(msg)

    if max_n_frames is not None and metadata.num_frames > max_n_frames:
        msg = f"Found file '{fpath}' already downloaded but it is invalid (audio is longer than {max_n_frames=} samples)."
        msgs.append(msg)

    if sr is not None and metadata.sample_rate != sr:
        msg = f"Found file '{fpath}' already downloaded but it is invalid (invalid {metadata.sample_rate=} != {sr})."
        msgs.append(msg)

    if n_channels is not None and metadata.num_channels != n_channels:
        msg = f"Found file '{fpath}' already downloaded but it is invalid (invalid {metadata.num_channels=} != {sr})."
        msgs.append(msg)

    if len(msgs) > 0:
        logger.error(f"Found {len(msgs)} error(s) when validating audio file:")
        for msg in msgs:
            logger.error(msg)

    return len(msgs) == 0


def _get_youtube_link(youtube_id: str, start_time: Optional[int]) -> str:
    link = f"https://www.youtube.com/watch?v={youtube_id}"
    if start_time is None:
        return link
    else:
        return f"{link}&t={start_time}s"


# Audio directory names per subset
_AUDIOCAPS_AUDIO_DNAMES: Dict[AudioCapsSubset, str] = {
    "train": "train",
    "val": "val",
    "test": "test",
    "train_fixed": "train",
}

# Internal typing to make easier to add new links without error
_AudioCapsLinkType = Literal["captions"]

# Archives and file links used to download AudioCaps labels and metadata
_AUDIOCAPS_LINKS: Dict[
    AudioCapsVersion, Dict[AudioCapsSubset, Dict[_AudioCapsLinkType, LinkInfo]]
] = {
    "v1": {
        "train": {
            "captions": {
                "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset/train.csv",
                "fname": "train.csv",
            },
        },
        "val": {
            "captions": {
                "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset/val.csv",
                "fname": "val.csv",
            },
        },
        "test": {
            "captions": {
                "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset/test.csv",
                "fname": "test.csv",
            },
        },
        "train_fixed": {
            "captions": {
                "url": "https://raw.githubusercontent.com/Labbeti/aac-datasets/dev/data/audiocaps/train_fixed.csv",
                "fname": "train_fixed.csv",
            },
        },
    },
    "v2": {
        "train": {
            "captions": {
                "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset2.0/train.csv",
                "fname": "train.csv",
            },
        },
        "val": {
            "captions": {
                "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset2.0/val.csv",
                "fname": "val.csv",
            },
        },
        "test": {
            "captions": {
                "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset2.0/test.csv",
                "fname": "test.csv",
            },
        },
    },
}

# Archives and file links used to download AudioSet metadata
_AUDIOSET_LINKS: Dict[str, LinkInfo] = {
    "class_labels_indices": {
        "fname": "class_labels_indices.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/class_labels_indices.csv",
    },
    "eval": {
        "fname": "eval_segments.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/eval_segments.csv",
    },
    "balanced": {
        "fname": "balanced_train_segments.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/balanced_train_segments.csv",
    },
    "unbalanced": {
        "fname": "unbalanced_train_segments.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/unbalanced_train_segments.csv",
    },
}

# Audio filename format for AudioCaps
_AUDIO_FNAME_FORMAT = "{youtube_id}_{start_time}.{audio_format}"

_AUDIOCAPS_OLD_SUBSETS_NAMES: Dict[str, AudioCapsSubset] = {
    "train_v2": "train_fixed",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datasets import DatasetInfo
from typing_extensions import TypedDict


class DatasetCard:
    def to_dataset_info(self) -> DatasetInfo:
        empty_info = DatasetInfo()
        info = {
            k.lower(): v
            for k, v in self.__dict__.items()
            if k.lower() in empty_info.__dict__
        }
        return DatasetInfo(**info)


class LinkInfo(TypedDict):
    fname: str
    url: str


class LinkInfoHash(TypedDict):
    fname: str
    url: str
    hash_value: str



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import logging
import os.path as osp
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import pythonwrench as pw
import torchaudio
import tqdm
from datasets import Dataset as HFDataset
from torch import Tensor
from torch.utils.data.dataset import Dataset
from torchwrench import IntegralTensor0D, IntegralTensor1D
from typing_extensions import TypeAlias, TypeGuard

from aac_datasets.datasets.functional.common import DatasetCard
from aac_datasets.utils.typing import AudioMetaData

logger = logging.getLogger(__name__)

ItemType = TypeVar("ItemType", covariant=True)
IndexType: TypeAlias = Union[int, Iterable[int], Iterable[bool], Tensor, slice, None]
ColumnType: TypeAlias = Union[str, Iterable[str], None]

_INDEX_TYPES = ("int", "Iterable[int]", "Iterable[bool]", "Tensor", "slice", "None")


class AACDataset(Generic[ItemType], Dataset[ItemType]):
    """Base class for AAC datasets."""

    # Initialization
    def __init__(
        self,
        raw_data: Optional[Dict[str, List[Any]]] = None,
        transform: Optional[Callable[[ItemType], Any]] = None,
        column_names: Optional[Iterable[str]] = None,
        flat_captions: bool = False,
        sr: Union[int, Iterable[int], None] = None,
        verbose: int = 0,
    ) -> None:
        if raw_data is None:
            raw_data = {}
        if column_names is None:
            column_names = raw_data.keys()
        column_names = list(column_names)
        if isinstance(sr, Iterable):
            sr = list(sr)

        if len(raw_data) > 1:
            size = len(next(iter(raw_data.values())))
            invalid_columns = [col for col, lst in raw_data.items() if len(lst) != size]
            if len(invalid_columns) > 0:
                msg = f"Invalid raw_data number of items in the following columns: {tuple(invalid_columns)}."
                raise ValueError(msg)

        super().__init__()
        self._raw_data = raw_data
        self._transform = transform
        self._columns = column_names
        self._flat_captions = flat_captions
        self._sr = sr
        self._verbose = verbose

        self._online_fns = {}
        self._sizes = []

        if self._flat_captions:
            self._flat_raw_data()

    @staticmethod
    def new_empty() -> "AACDataset":
        """Create a new empty dataset."""
        return AACDataset(
            raw_data={},
            transform=None,
            column_names=(),
            flat_captions=False,
            sr=None,
            verbose=0,
        )

    # Properties
    @property
    def all_columns(self) -> List[str]:
        """The name of all columns of the dataset."""
        return list(pw.union_dicts([self._raw_data, self._online_fns]))

    @property
    def column_names(self) -> List[str]:
        """The name of all selected column of the dataset."""
        return self._columns

    @property
    def flat_captions(self) -> bool:
        """Returns true if captions has been flattened."""
        return self._flat_captions

    @property
    def num_columns(self) -> int:
        """Number of columns in the dataset."""
        return len(self.column_names)

    @property
    def num_rows(self) -> int:
        """Number of rows in the dataset (same as len())."""
        return len(self)

    @property
    def raw_data(self) -> Dict[str, List[Any]]:
        return self._raw_data

    @property
    def shape(self) -> Tuple[int, int]:
        """Shape of the dataset (number of columns, number of rows)."""
        return len(self), len(self.column_names)

    @property
    def sr(self) -> Union[int, List[int], None]:
        return self._sr

    @property
    def transform(self) -> Optional[Callable]:
        return self._transform

    @property
    def verbose(self) -> int:
        return self._verbose

    @column_names.setter
    def column_names(
        self,
        columns: Iterable[str],
    ) -> None:
        columns = list(columns)
        self._check_columns(columns)
        self._columns = columns

    @transform.setter
    def transform(self, transform: Optional[Callable[[ItemType], Any]]) -> None:
        self._transform = transform

    @verbose.setter
    def verbose(self, verbose: int) -> None:
        self._verbose = verbose

    # Public methods
    @pw.deprecated_function()
    def at(self, *args, **kwargs) -> Any:
        """Deprecated: Use get_item method instead."""
        return self.get_item(*args, **kwargs)

    @overload
    def get_item(self, index: int) -> ItemType: ...

    @overload
    def get_item(  # type: ignore
        self,
        index: Union[Iterable[int], Iterable[bool], slice, None],
        column: str,
    ) -> List: ...

    @overload
    def get_item(
        self,
        index: Union[Iterable[int], Iterable[bool], slice, None],
        column: Union[Iterable[str], None] = None,
    ) -> Dict[str, List]: ...

    @overload
    def get_item(self, index: IndexType, column: ColumnType) -> Any: ...

    def get_item(
        self,
        index: IndexType = None,
        column: ColumnType = None,
    ) -> Any:
        """Get a specific data field.

        :param index: The index or slice of the value in range [0, len(dataset)-1].
        :param column: The name(s) of the column. Can be any value of :meth:`~Clotho.columns`.
        :returns: The field value. The type depends of the column.
        """
        if index is None:
            index = slice(None)

        elif isinstance(index, Tensor):
            if not pw.isinstance_generic(index, (IntegralTensor0D, IntegralTensor0D)):
                msg = "Invalid tensor dtype. (expected integral 0d or 1d tensor)"
                raise TypeError(msg)

            index = index.tolist()

        if column is None:
            column = self.column_names

        if not isinstance(column, str) and isinstance(column, Iterable):
            return {column_i: self.get_item(index, column_i) for column_i in column}

        if isinstance(index, (int, slice)) and (
            column in self._raw_data.keys() and column not in self._online_fns
        ):
            return self._raw_data[column][index]  # type: ignore

        if isinstance(index, slice):
            index = range(len(self))[index]

        if isinstance(index, Iterable):
            index = list(index)
            if pw.isinstance_generic(index, Iterable[bool]):
                if len(index) != len(self):
                    msg = f"The length of the mask ({len(index)}) does not match the length of the dataset ({len(self)})."
                    raise IndexError(msg)
                index = [i for i, idx_i in enumerate(index) if idx_i]

            elif not pw.isinstance_generic(index, Iterable[int]):
                msg = f"Invalid input type for {index=}. (expected Iterable[int], not Iterable[{index[0].__class__.__name__}])"
                raise TypeError(msg)

            values = [
                self.get_item(idx_i, column)
                for idx_i in tqdm.tqdm(
                    index,
                    desc=f"Loading column '{column}'...",
                    disable=self._verbose < 2,
                )
            ]
            return values

        if not isinstance(index, int):
            msg = (
                f"Invalid argument type {type(index)}. (expected one of {_INDEX_TYPES})"
            )
            raise TypeError(msg)

        return self._load_online_value(column, index)

    def has_raw_column(self, column: str) -> bool:
        """Returns True if column name exists in raw data."""
        return column in self._raw_data

    def has_post_column(self, column: str) -> bool:
        """Returns True if column name exists in post processed data."""
        return column in self._online_fns

    def has_column(self, column: str) -> bool:
        """Returns True if column name exists in data."""
        return self.has_raw_column(column) or self.has_post_column(column)

    def remove_column(self, column: str) -> Union[List[Any], Callable]:
        """Removes a column from this dataset."""
        if column in self._raw_data:
            column_data = self._raw_data.pop(column, [])
            return column_data
        elif column in self._online_fns:
            fn = self._online_fns.pop(column)
            return fn
        else:
            raise ValueError(f"Column '{column}' does not exists in dataset.")

    def rename_column(
        self,
        old_column: str,
        new_column: str,
        allow_replace: bool = False,
    ) -> None:
        """Renames a column from this dataset."""
        column_data_or_fn = self.remove_column(old_column)

        if isinstance(column_data_or_fn, List):
            self.add_raw_column(new_column, column_data_or_fn, allow_replace)
        elif isinstance(column_data_or_fn, Callable):
            self.add_online_column(new_column, column_data_or_fn, allow_replace)
        else:
            msg = f"Invalid type {type(column_data_or_fn)}. (expected List or Callable)"
            raise TypeError(msg)

    def add_raw_column(
        self,
        column_name: str,
        column_data: List[Any],
        allow_replace: bool = False,
    ) -> None:
        """Add a new raw column to this dataset."""
        if not allow_replace and column_name in self._raw_data:
            msg = f"Column '{column_name}' already exists. Please choose another name or set allow_replace arg to True."
            raise ValueError(msg)
        if len(self._raw_data) > 0 and len(column_data) != len(self):
            msg = f"Invalid number of rows in column '{column_name}'."
            raise ValueError(msg)
        self._raw_data[column_name] = column_data

    def add_online_column(
        self,
        column: str,
        load_fn: Callable[[Any, int], Any],
        allow_replace: bool = False,
    ) -> None:
        """Add a new post-processed column to this dataset."""
        if not allow_replace and column in self._online_fns:
            msg = f"Column '{column}' already exists in {self} and found argument {allow_replace=}."
            raise ValueError(msg)
        self._online_fns[column] = load_fn

    def add_online_columns(
        self,
        post_columns_fns: Dict[str, Callable[[Any, int], Any]],
        allow_replace: bool = False,
    ) -> None:
        """Add several new post-processed columns to this dataset."""
        for name, load_fn in post_columns_fns.items():
            self.add_online_column(name, load_fn, allow_replace)

    def preload_online_column(
        self,
        column: str,
        allow_replace: bool = False,
    ) -> Callable[[Any, int], Any]:
        """Load all data from a post-column data into raw data."""
        if column not in self._online_fns:
            msg = f"Invalid argument {column=}."
            raise ValueError(msg)

        column_data = [
            self._load_online_value(column, i)
            for i in tqdm.trange(
                len(self),
                disable=self._verbose < 2,
                desc=f"Preloading column '{column}'",
            )
        ]
        fn = self._online_fns.pop(column)
        self.add_raw_column(column, column_data, allow_replace=allow_replace)
        return fn

    def to_dict(self, load_online_values: bool = False) -> Dict[str, List[Any]]:
        """Convert dataset to dictionary.

        :param load_online_values: If True, load ALL online values (e.g. audio waveform). Otherwise load only the raw data of the dataset. defaults to False.
        """
        raw_data = copy.copy(self._raw_data)
        if load_online_values:
            for column_name in self._online_fns.keys():
                column_data = self.get_item(None, column_name)
                raw_data[column_name] = column_data
        return raw_data

    def to_list(self, load_online_values: bool = False) -> List[ItemType]:
        """Convert dataset to list.

        :param load_online_values: If True, load ALL online values (e.g. audio waveform). Otherwise load only the raw data of the dataset. defaults to False.
        """
        raw_data = self.to_dict(load_online_values)
        return pw.dict_list_to_list_dict(raw_data, key_mode="same")  # type: ignore

    def to_hf_dataset(self, load_online_values: bool = False) -> HFDataset:
        datadict = self.to_dict(load_online_values=False)

        dataset_info = None
        if hasattr(self.__class__, "CARD"):
            card = self.__class__.CARD  # type: ignore
            if isinstance(card, DatasetCard):
                dataset_info = card.to_dataset_info()

        split = getattr(self, "subset", None)

        hf_dataset = HFDataset.from_dict(datadict, info=dataset_info, split=split)
        del datadict

        if load_online_values:
            for colname, fn in self._online_fns.items():

                def wrapped_fn(index: int) -> Any:
                    return {colname: fn(self, index)}

                hf_dataset = hf_dataset.map(
                    wrapped_fn,
                    input_columns="index",
                    load_from_cache_file=False,
                    keep_in_memory=False,
                )

        return hf_dataset

    # Magic methods
    @overload
    def __getitem__(self, index: int) -> ItemType: ...

    @overload
    def __getitem__(self, index: Tuple[Union[Iterable[int], slice, None], str]) -> List:  # type: ignore
        ...

    @overload
    def __getitem__(
        self, index: Union[Iterable[int], slice, None]
    ) -> Dict[str, List]: ...

    @overload
    def __getitem__(
        self,
        index: Tuple[Union[Iterable[int], slice, None], Union[Iterable[str], None]],
    ) -> Dict[str, List]: ...

    @overload
    def __getitem__(self, index: Any) -> Any: ...

    def __getitem__(self, index: Union[IndexType, Tuple[IndexType, ColumnType]]) -> Any:  # type: ignore
        if (
            isinstance(index, tuple)
            and len(index) == 2
            and _is_index(index[0])
            and _is_column(index[1])
        ):
            index, column = index
        else:
            column = None

        item = self.get_item(index, column)  # type: ignore

        if (
            isinstance(index, int)
            and self._transform is not None
            and (
                column is None
                or (
                    isinstance(column, Iterable)
                    and not isinstance(column, str)
                    and set(column) == set(self._columns)
                )
            )
        ):
            item = self._transform(item)  # type: ignore
        return item

    def __len__(self) -> int:
        """
        :return: The number of items in the dataset.
        """
        if len(self._raw_data) > 0:
            return len(next(iter(self._raw_data.values())))
        else:
            return 0

    def __repr__(self) -> str:
        info = {
            "size": len(self),
            "num_columns": self.num_columns,
        }
        repr_str = ", ".join(f"{k}={v}" for k, v in info.items())
        return f"{self.__class__.__name__}({repr_str})"

    # Private methods
    def _check_columns(self, columns: List[str]) -> None:
        expected_columns = dict.fromkeys(self.all_columns)
        invalid_columns = [name for name in columns if name not in expected_columns]
        if len(invalid_columns) > 0:
            msg = f"Invalid argument {columns=}. (found {len(invalid_columns)} invalids column names for {self.__class__.__name__}: {invalid_columns})"
            raise ValueError(msg)

        invalid_columns = [name for name in columns if not self.has_column(name)]
        if len(invalid_columns) > 0:
            msg = f"Invalid argument {columns=}. (found {len(invalid_columns)} invalids column names for {self.__class__.__name__}: {invalid_columns})"
            raise ValueError(msg)

    def _flat_raw_data(self) -> None:
        raw_data, sizes = _flat_raw_data(self._raw_data)
        self._raw_data = raw_data
        self._sizes = sizes

    def _unflat_raw_data(self) -> None:
        raw_data = _unflat_raw_data(self._raw_data, self._sizes)
        self._raw_data = raw_data

    def _load_online_value(self, column: str, index: int) -> Any:
        if column in self._online_fns:
            fn = self._online_fns[column]
            return fn(self, index)
        else:
            msg = f"Invalid argument column={column} at {index=}. (expected one of {self.all_columns})"
            raise ValueError(msg)

    def _load_audio(self, index: int) -> Tensor:
        fpath = self.get_item(index, "fpath")
        audio_and_sr: Tuple[Tensor, int] = torchaudio.load(fpath)  # type: ignore
        audio, sr = audio_and_sr

        # Sanity check
        if audio.nelement() == 0:
            msg = f"Invalid audio number of elements in {fpath}. (expected {audio.nelement()=} > 0)"
            raise RuntimeError(msg)

        if self._sr is not None and (self._sr != sr):
            msg = (
                f"Invalid sample rate {sr}Hz for audio {fpath}. (expected {self._sr}Hz)"
            )
            raise RuntimeError(msg)
        return audio

    def _load_audio_metadata(self, index: int) -> AudioMetaData:
        fpath = self.get_item(index, "fpath")
        audio_metadata = torchaudio.info(fpath)  # type: ignore
        return audio_metadata

    def _load_duration(self, index: int) -> float:
        audio_metadata: AudioMetaData = self.get_item(index, "audio_metadata")
        duration = audio_metadata.num_frames / audio_metadata.sample_rate
        return duration

    def _load_fname(self, index: int) -> str:
        fpath = self.get_item(index, "fpath")
        fname = osp.basename(fpath)
        return fname

    def _load_num_channels(self, index: int) -> int:
        audio_metadata = self.get_item(index, "audio_metadata")
        num_channels = audio_metadata.num_channels
        return num_channels

    def _load_num_frames(self, index: int) -> int:
        audio_metadata = self.get_item(index, "audio_metadata")
        num_frames = audio_metadata.num_frames
        return num_frames

    def _load_sr(self, index: int) -> int:
        audio_metadata = self.get_item(index, "audio_metadata")
        sr = audio_metadata.sample_rate
        return sr


def _is_index(index: Any) -> TypeGuard[IndexType]:
    return pw.isinstance_generic(
        index,
        (
            int,
            Iterable[int],
            Iterable[bool],
            slice,
            pw.NoneType,
            IntegralTensor0D,
            IntegralTensor1D,
        ),
    )


def _is_column(column: Any) -> TypeGuard[ColumnType]:
    return pw.isinstance_generic(column, (Iterable[str], pw.NoneType))


def _flat_raw_data(
    raw_data: Dict[str, List[Any]],
    caps_column: str = "captions",
) -> Tuple[Dict[str, List[Any]], List[int]]:
    if caps_column not in raw_data:
        msg = f"Cannot flat raw data without '{caps_column}' column. (found only columns {tuple(raw_data.keys())})"
        raise ValueError(msg)

    mcaps: List[List[str]] = raw_data[caps_column]
    raw_data_flat = {key: [] for key in raw_data.keys()}

    for i, caps in enumerate(mcaps):
        if len(caps) == 0:
            for key in raw_data.keys():
                raw_data_flat[key].append(raw_data[key][i])
        else:
            for cap in caps:
                for key in raw_data.keys():
                    if key == caps_column:
                        continue
                    raw_data_flat[key].append(raw_data[key][i])

                # Overwrite cap
                raw_data_flat[caps_column].append([cap])

    sizes = [len(caps) for caps in mcaps]
    return raw_data_flat, sizes


def _unflat_raw_data(
    raw_data_flat: Dict[str, List[Any]],
    sizes: List[int],
    caps_column: str = "captions",
) -> Dict[str, List[Any]]:
    if caps_column not in raw_data_flat:
        msg = f"Cannot flat raw data without '{caps_column}' column. (found only columns {tuple(raw_data_flat.keys())})"
        raise ValueError(msg)

    raw_data = {key: [] for key in raw_data_flat.keys()}

    cumsize = 0
    for size in sizes:
        for key in raw_data.keys():
            if key == caps_column:
                caps = [
                    raw_data_flat[key][index][0]
                    for index in range(cumsize, cumsize + size)
                ]
                raw_data[key].append(caps)
            else:
                raw_data[key].append(raw_data_flat[key][cumsize])
        cumsize += size

    return raw_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import os
import os.path as osp
from pathlib import Path
from typing import Dict, Union

from torch.hub import download_url_to_file

_AUDIOSET_INFOS = {
    "class_labels_indices": {
        "fname": "class_labels_indices.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/class_labels_indices.csv",
    },
}
_DEFAULT_CACHE_PATH = Path.home().joinpath(".cache", "audioset_mapping")


def get_audioset_mapping_cache_path(cache_path: Union[str, Path, None] = None) -> Path:
    if cache_path is not None:
        return Path(cache_path)
    else:
        return _DEFAULT_CACHE_PATH


def download_audioset_mapping(
    cache_path: Union[str, Path, None] = None,
    verbose: int = 0,
) -> None:
    cache_path = get_audioset_mapping_cache_path(cache_path)
    os.makedirs(cache_path, exist_ok=True)

    info = _AUDIOSET_INFOS["class_labels_indices"]
    map_fname = info["fname"]
    map_fpath = cache_path.joinpath(map_fname)

    url = info["url"]
    download_url_to_file(url, str(map_fpath), progress=verbose >= 1)


def load_audioset_mapping(
    key_name: str = "index",
    val_name: str = "display_name",
    offline: bool = False,
    cache_path: Union[str, Path, None] = None,
    verbose: int = 0,
) -> Dict:
    NAMES = ("index", "mid", "display_name")
    if key_name not in NAMES:
        raise ValueError(f"Invalid argument {key_name=}. (expected one of {NAMES})")
    if val_name not in NAMES:
        raise ValueError(f"Invalid argument {val_name=}. (expected one of {NAMES})")
    if key_name == val_name:
        raise ValueError(
            f"Invalid arguments key_name={key_name} with {val_name=}. (expected different values)"
        )

    cache_path = get_audioset_mapping_cache_path(cache_path)

    info = _AUDIOSET_INFOS["class_labels_indices"]
    map_fname = info["fname"]
    map_fpath = cache_path.joinpath(map_fname)

    if not osp.isfile(map_fpath):
        if offline:
            msg = f"Cannot find or download audioset mapping file in '{map_fpath}' with mode {offline=}."
            raise FileNotFoundError(msg)

        download_audioset_mapping(cache_path, verbose)

    with open(map_fpath, "r") as file:
        reader = csv.DictReader(file, skipinitialspace=True, strict=True)
        data = list(reader)

    keys = [data_i[key_name] for data_i in data]
    values = [data_i[val_name] for data_i in data]

    if key_name == "index":
        keys = list(map(int, keys))
    if val_name == "index":
        values = list(map(int, values))

    mapping = dict(zip(keys, values))
    return mapping


def load_audioset_name_to_idx(
    offline: bool = False,
    cache_path: Union[str, Path, None] = None,
    verbose: int = 0,
) -> Dict[str, int]:
    return load_audioset_mapping("display_name", "index", offline, cache_path, verbose)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import logging
import os.path as osp
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict, List, Optional, Union

from torch import Tensor
from typing_extensions import TypedDict

from aac_datasets.datasets.base import AACDataset
from aac_datasets.datasets.functional.macs import (
    MACSCard,
    MACSSubset,
    _get_audio_dpath,
    download_macs_dataset,
    load_macs_dataset,
)
from aac_datasets.utils.globals import _get_root

logger = logging.getLogger(__name__)


class MACSItem(TypedDict):
    r"""Dataclass representing a single MACS item."""

    # Common attributes
    audio: Tensor
    captions: List[str]
    dataset: str
    fname: str
    index: int
    subset: MACSSubset
    sr: int
    duration: float
    # MACS-specific attributes
    annotators_ids: List[str]
    competences: List[float]
    identifier: str
    scene_label: str
    tags: List[List[str]]


class MACS(AACDataset[MACSItem]):
    r"""Unofficial MACS PyTorch dataset.

    .. code-block:: text
        :caption: Dataset folder tree

        {root}
        └── MACS
            ├── audio
            │    └── (3930 wav files, ~13GB)
            ├── LICENCE.txt
            ├── MACS.yaml
            ├── MACS_competence.csv
            └── tau_meta
                ├── fold1_evaluate.csv
                ├── fold1_test.csv
                ├── fold1_train.csv
                └── meta.csv
    """

    # Common globals
    CARD: ClassVar[MACSCard] = MACSCard()

    # Initialization
    def __init__(
        self,
        # Common args
        root: Union[str, Path, None] = None,
        subset: MACSSubset = MACSCard.DEFAULT_SUBSET,
        download: bool = False,
        transform: Optional[Callable[[MACSItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        *,
        # MACS-specific args
        clean_archives: bool = True,
        flat_captions: bool = False,
    ) -> None:
        """
        :param root: The parent of the dataset root directory.
            The data will be stored in the 'MACS' subdirectory.
            defaults to ".".
        :param subset: The subset of the dataset. This parameter is here only to accept the same interface than the other datasets.
            The only valid subset is "full" and other values will raise a ValueError.
            defaults to "full".
        :param download: Download the dataset if download=True and if the dataset is not already downloaded.
            defaults to False.
        :param transform: The transform to apply to the global dict item. This transform is applied only in getitem method when argument is an integer.
            defaults to None.
        :param verbose: Verbose level to use. Can be 0 or 1.
            defaults to 0.
        :param force_download: If True, force to re-download file even if they exists on disk.
            defaults to False.
        :param verify_files: If True, check hash value when possible.
            defaults to False.

        :param clean_archives: If True, remove the compressed archives from disk to save space.
            defaults to True.
        :param flat_captions: If True, map captions to audio instead of audio to caption.
            defaults to True.
        """
        if subset not in MACSCard.SUBSETS:
            msg = f"Invalid argument {subset=} for MACS. (expected one of {MACSCard.SUBSETS})"
            raise ValueError(msg)

        root = _get_root(root)

        if download:
            download_macs_dataset(
                root=root,
                subset=subset,
                force=force_download,
                verbose=verbose,
                clean_archives=clean_archives,
                verify_files=verify_files,
            )

        raw_data, annotator_id_to_competence = load_macs_dataset(
            root=root,
            subset=subset,
            verbose=verbose,
        )

        audio_dpath = _get_audio_dpath(root)
        size = len(next(iter(raw_data.values())))
        raw_data["dataset"] = [MACSCard.NAME] * size
        raw_data["subset"] = [subset] * size
        raw_data["fpath"] = [
            osp.join(audio_dpath, fname) for fname in raw_data["fname"]
        ]
        raw_data["index"] = list(range(size))

        column_names = list(MACSItem.__required_keys__) + list(  # type: ignore
            MACSItem.__optional_keys__  # type: ignore
        )
        super().__init__(
            raw_data=raw_data,
            transform=transform,
            column_names=column_names,
            flat_captions=flat_captions,
            sr=MACSCard.SAMPLE_RATE,
            verbose=verbose,
        )
        self._root = root
        self._subset: MACSSubset = subset
        self._download = download
        self._transform = transform
        self._flat_captions = flat_captions
        self._verbose = verbose
        self._annotator_id_to_competence = annotator_id_to_competence

        self.add_online_columns(
            {
                "audio": MACS._load_audio,
                "audio_metadata": MACS._load_audio_metadata,
                "duration": MACS._load_duration,
                "num_channels": MACS._load_num_channels,
                "num_frames": MACS._load_num_frames,
                "sr": MACS._load_sr,
                "competences": MACS._load_competences,
            }
        )

    # Properties
    @property
    def download(self) -> bool:
        return self._download

    @property
    def root(self) -> str:
        return self._root

    @property
    def sr(self) -> int:
        return self._sr  # type: ignore

    @property
    def subset(self) -> MACSSubset:
        return self._subset

    # Public methods
    def get_annotator_id_to_competence_dict(self) -> Dict[int, float]:
        """Get annotator to competence dictionary."""
        # Note : copy to prevent any changes on this attribute
        return copy.deepcopy(self._annotator_id_to_competence)

    def get_competence(self, annotator_id: int) -> float:
        """Get competence value for a specific annotator id."""
        return self._annotator_id_to_competence[annotator_id]

    def _load_competences(self, index: int) -> List[float]:
        annotators_ids: List[int] = self.get_item(index, "annotators_ids")
        competences = [self.get_competence(id_) for id_ in annotators_ids]
        return competences

    # Magic methods
    def __repr__(self) -> str:
        repr_dic = {
            "subset": self._subset,
            "size": len(self),
            "num_columns": len(self.column_names),
        }
        repr_str = ", ".join(f"{k}={v}" for k, v in repr_dic.items())
        return f"{MACSCard.PRETTY_NAME}({repr_str})"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from argparse import ArgumentParser, Namespace

import pythonwrench as pw
import yaml

from aac_datasets.datasets.functional.audiocaps import (
    AudioCapsCard,
    download_audiocaps_datasets,
)
from aac_datasets.datasets.functional.clotho import ClothoCard, download_clotho_datasets
from aac_datasets.datasets.functional.macs import MACSCard, download_macs_datasets
from aac_datasets.datasets.functional.wavcaps import (
    WavCapsCard,
    download_wavcaps_datasets,
)
from aac_datasets.utils.globals import (
    get_default_ffmpeg_path,
    get_default_root,
    get_default_ytdlp_path,
    get_default_zip_path,
)

logger = logging.getLogger(__name__)


def _get_main_download_args() -> Namespace:
    parser = ArgumentParser(
        description="Download a dataset at specified root directory.",
    )

    parser.add_argument(
        "--root",
        type=str,
        default=get_default_root(),
        help="The path to the parent directory of the datasets.",
    )
    parser.add_argument(
        "--force",
        type=pw.str_to_bool,
        default=False,
        help="Force download of files, even if they are already downloaded.",
    )
    parser.add_argument(
        "--verbose",
        type=int,
        default=1,
        help="Verbose level of the script. 0 means silent mode, 1 is default mode and 2 add additional debugging outputs.",
    )

    subparsers = parser.add_subparsers(
        dest="dataset",
        required=True,
        description="The dataset to download.",
    )

    audiocaps_subparser = subparsers.add_parser(AudioCapsCard.NAME)
    audiocaps_subparser.add_argument(
        "--ffmpeg_path",
        type=str,
        default=get_default_ffmpeg_path(),
        help="Path to ffmpeg used to download audio from youtube.",
    )
    audiocaps_subparser.add_argument(
        "--ytdlp_path",
        type=str,
        default=get_default_ytdlp_path(),
        help="Path to yt-dl program used to extract metadata from a youtube video.",
    )
    audiocaps_subparser.add_argument(
        "--ytdlp_opts",
        type=str,
        nargs="+",
        default=[],
        help="yt-dlp optional arguments.",
    )
    audiocaps_subparser.add_argument(
        "--with_tags",
        type=pw.str_to_bool,
        default=True,
        help="Download additional audioset tags corresponding to audiocaps audio.",
    )
    audiocaps_subparser.add_argument(
        "--subsets",
        type=str,
        default=AudioCapsCard.SUBSETS,
        nargs="+",
        choices=AudioCapsCard.SUBSETS,
        help="AudioCaps subsets to download.",
    )
    audiocaps_subparser.add_argument(
        "--max_workers",
        type=pw.str_to_optional_int,
        default=1,
        help="Number of workers used for downloading multiple files in parallel.",
    )
    audiocaps_subparser.add_argument(
        "--version",
        type=str,
        default=AudioCapsCard.DEFAULT_VERSION,
        choices=AudioCapsCard.VERSIONS,
        help="The version of the AudioCaps dataset.",
    )
    audiocaps_subparser.add_argument(
        "--num_dl_attempts",
        type=int,
        default=2,
        help="Number of download attempts.",
    )

    clotho_subparser = subparsers.add_parser(ClothoCard.NAME)
    clotho_subparser.add_argument(
        "--version",
        type=str,
        default=ClothoCard.DEFAULT_VERSION,
        choices=ClothoCard.VERSIONS,
        help="The version of the Clotho dataset.",
    )
    clotho_subparser.add_argument(
        "--clean_archives",
        type=pw.str_to_bool,
        default=False,
        help="Remove archives files after extraction.",
    )
    clotho_subparser.add_argument(
        "--subsets",
        type=str,
        default=ClothoCard.SUBSETS,
        nargs="+",
        choices=ClothoCard.SUBSETS,
        help="Clotho subsets to download. Available subsets depends of the Clotho version used.",
    )

    macs_subparser = subparsers.add_parser(MACSCard.NAME)
    # Note : MACS only have 1 subset, so we do not add MACS subsets arg
    macs_subparser.add_argument(
        "--clean_archives",
        type=pw.str_to_bool,
        default=False,
        help="Remove archives files after extraction.",
    )
    macs_subparser.add_argument(
        "--verify_files",
        type=pw.str_to_bool,
        default=True,
        help="Verify if downloaded files have a valid checksum.",
    )

    wavcaps_subparser = subparsers.add_parser(WavCapsCard.NAME)
    wavcaps_subparser.add_argument(
        "--clean_archives",
        type=pw.str_to_bool,
        default=False,
        help="Remove archives files after extraction.",
    )
    wavcaps_subparser.add_argument(
        "--subsets",
        type=str,
        default=WavCapsCard.SUBSETS,
        nargs="+",
        choices=WavCapsCard.SUBSETS,
        help="WavCaps subsets to download.",
    )
    wavcaps_subparser.add_argument(
        "--hf_cache_dir",
        type=pw.str_to_optional_str,
        default=None,
        help="Hugging face cache dir.",
    )
    wavcaps_subparser.add_argument(
        "--revision",
        type=str,
        default=WavCapsCard.DEFAULT_REVISION,
        help="Revision of the WavCaps dataset.",
    )
    wavcaps_subparser.add_argument(
        "--zip_path",
        type=str,
        default=get_default_zip_path(),
        help="Path to zip executable to combine and extract WavCaps archives.",
    )

    args = parser.parse_args()
    return args


def _main_download() -> None:
    args = _get_main_download_args()
    pw.setup_logging_verbose("aac_datasets", args.verbose)

    if args.verbose >= 2:
        logger.debug(yaml.dump({"Arguments": args.__dict__}, sort_keys=False))

    if args.dataset == AudioCapsCard.NAME:
        download_audiocaps_datasets(
            root=args.root,
            subsets=args.subsets,
            force=args.force,
            verbose=args.verbose,
            ffmpeg_path=args.ffmpeg_path,
            max_workers=args.max_workers,
            with_tags=args.with_tags,
            ytdlp_path=args.ytdlp_path,
            ytdlp_opts=args.ytdlp_opts,
            version=args.version,
            num_dl_attempts=args.num_dl_attempts,
        )

    elif args.dataset == ClothoCard.NAME:
        download_clotho_datasets(
            root=args.root,
            subsets=args.subsets,
            force=args.force,
            verbose=args.verbose,
            clean_archives=args.clean_archives,
            version=args.version,
        )

    elif args.dataset == MACSCard.NAME:
        download_macs_datasets(
            root=args.root,
            force=args.force,
            verbose=args.verbose,
            clean_archives=args.clean_archives,
        )

    elif args.dataset == WavCapsCard.NAME:
        download_wavcaps_datasets(
            root=args.root,
            subsets=args.subsets,
            force=args.force,
            verbose=args.verbose,
            clean_archives=args.clean_archives,
            hf_cache_dir=args.hf_cache_dir,
            revision=args.revision,
            zip_path=args.zip_path,
        )

    else:
        DATASET_NAMES = (
            AudioCapsCard.NAME,
            ClothoCard.NAME,
            MACSCard.NAME,
            WavCapsCard.NAME,
        )
        msg = f"Invalid argument {args.dataset}. (expected one of {DATASET_NAMES})"
        raise ValueError(msg)


if __name__ == "__main__":
    _main_download()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


def _print_usage() -> None:
    print(
        "Command line usage :\n"
        "- Download a dataset             : aac-datasets-download [--root ROOT] [--verbose VERBOSE] [--force (false|true)] (clotho|audiocaps|macs) [ARGS...]\n"
        "- Check a installation directory : aac-datasets-check [--root ROOT] [--verbose VERBOSE]\n"
        "- Print package install info     : aac-datasets-info\n"
        "- Show this usage page           : aac-datasets\n"
    )


if __name__ == "__main__":
    _print_usage()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import logging
import os
import os.path as osp
import shutil
import zipfile
from pathlib import Path
from typing import Any, Dict, Iterable, List, Tuple, Union, get_args

import yaml
from torchwrench.hub.download import download_file, hash_file
from typing_extensions import Literal

from aac_datasets.datasets.functional.common import DatasetCard, LinkInfoHash
from aac_datasets.utils.globals import _get_root

logger = logging.getLogger(__name__)


MACSSubset = Literal["full"]


class MACSCard(DatasetCard):
    ANNOTATIONS_CREATORS: Tuple[str, ...] = ("crowdsourced",)
    CITATION: str = r"""
    @inproceedings{Martin2021b,
        title        = {Diversity and Bias in Audio Captioning Datasets},
        author       = {Martin, Irene and Mesaros, Annamaria},
        year         = 2021,
        month        = {November},
        booktitle    = {Proceedings of the 6th Detection and Classification of Acoustic Scenes and Events 2021 Workshop (DCASE2021)},
        address      = {Barcelona, Spain},
        pages        = {90--94},
        isbn         = {978-84-09-36072-7},
        url          = {https://dcase.community/documents/workshop2021/proceedings/DCASE2021Workshop_Martin_34.pdf},
        abstract     = {Describing soundscapes in sentences allows better understanding of the acoustic scene than a single label indicating the acoustic scene class or a set of audio tags indicating the sound events active in the audio clip. In addition, the richness of natural language allows a range of possible descriptions for the same acoustic scene. In this work, we address the diversity obtained when collecting descriptions of soundscapes using crowdsourcing. We study how much the collection of audio captions can be guided by the instructions given in the annotation task, by analysing the possible bias introduced by auxiliary information provided in the annotation process. Our study shows that even when given hints on the audio content, different annotators describe the same soundscape using different vocabulary. In automatic captioning, hints provided as audio tags represent grounding textual information that facilitates guiding the captioning output towards specific concepts. We also release a new dataset of audio captions and audio tags produced by multiple annotators for a subset of the TAU Urban Acoustic Scenes 2018 dataset, suitable for studying guided captioning.},
        doi.         = {10.5281/zenodo.5770113}
    }
    """
    DEFAULT_SUBSET: MACSSubset = "full"
    DESCRIPTION: str = "Multi-Annotator Captioned Soundscapes dataset."
    HOMEPAGE: str = "https://zenodo.org/record/5114771"
    LANGUAGE: Tuple[str, ...] = ("en",)
    LANGUAGE_DETAILS: Tuple[str, ...] = ("en-US",)
    MAX_CAPTIONS_PER_AUDIO: Dict[str, int] = {"full": 5}
    MIN_CAPTIONS_PER_AUDIO: Dict[str, int] = {"full": 2}
    NAME: str = "macs"
    N_CHANNELS: int = 2
    PRETTY_NAME: str = "MACS"
    SAMPLE_RATE: int = 48_000  # Hz
    SIZE_CATEGORIES: Tuple[str, ...] = ("1K<n<10K",)
    SUBSETS: Tuple[MACSSubset, ...] = get_args(MACSSubset)
    TASK_CATEGORIES: Tuple[str, ...] = ("audio-to-text", "text-to-audio")


def load_macs_dataset(
    # Common args
    root: Union[str, Path, None] = None,
    subset: MACSSubset = MACSCard.DEFAULT_SUBSET,
    verbose: int = 0,
) -> Tuple[Dict[str, List[Any]], Dict[int, float]]:
    """Load MACS metadata.

    :param root: Dataset root directory.
        defaults to ".".
    :param subset: The subset of MACS to use. Can be one of :attr:`~MACSCard.SUBSETS`.
        defaults to "full".
    :param verbose: Verbose level.
        defaults to 0.
    :returns: A dictionnary of lists containing each metadata.
    """

    root = _get_root(root)
    if not _is_prepared_macs(root):
        msg = f"Cannot load data: macs is not prepared in data {root=}. Please use download=True in dataset constructor."
        raise RuntimeError(msg)

    macs_dpath = _get_macs_root(root)
    tau_meta_dpath = _get_tau_meta_dpath(root)

    # Read data files
    captions_fname = MACS_FILES["captions"]["fname"]
    captions_fpath = osp.join(macs_dpath, captions_fname)
    if verbose >= 2:
        logger.debug(f"Reading captions file {captions_fname}...")

    with open(captions_fpath, "r") as file:
        caps_data = yaml.safe_load(file)

    tau_meta_fname = "meta.csv"
    tau_meta_fpath = osp.join(tau_meta_dpath, tau_meta_fname)
    if verbose >= 2:
        logger.debug(f"Reading Tau Urban acoustic scene meta file {tau_meta_fname}...")

    with open(tau_meta_fpath, "r") as file:
        reader = csv.DictReader(file, delimiter="\t")
        tau_tags_data = list(reader)

    competence_fname = MACS_FILES["annotators_competences"]["fname"]
    competence_fpath = osp.join(macs_dpath, competence_fname)
    if verbose >= 2:
        logger.debug(f"Reading file {competence_fname}...")

    with open(competence_fpath, "r") as file:
        reader = csv.DictReader(file, delimiter="\t")
        competences_data = list(reader)

    # Store MACS data
    raw_data: Dict[str, List[Any]] = {
        "fname": [item["filename"] for item in caps_data["files"]],
        "captions": [
            [subitem["sentence"] for subitem in item["annotations"]]
            for item in caps_data["files"]
        ],
        "tags": [
            [subitem["tags"] for subitem in item["annotations"]]
            for item in caps_data["files"]
        ],
        "annotators_ids": [
            [subitem["annotator_id"] for subitem in item["annotations"]]
            for item in caps_data["files"]
        ],
    }
    dataset_size = len(raw_data["fname"])

    # Build global mappings
    fname_to_idx = {fname: i for i, fname in enumerate(raw_data["fname"])}
    annotator_id_to_competence = {
        int(annotator["annotator_id"]): float(annotator["competence"])
        for annotator in competences_data
    }

    # Store TAU Urban acoustic scenes data
    tau_additional_keys = ("scene_label", "identifier")
    raw_data.update(
        {key: [None for _ in range(dataset_size)] for key in tau_additional_keys}
    )

    tau_meta_fpath = osp.join(tau_meta_dpath, "meta.csv")
    for tau_tags in tau_tags_data:
        fname = osp.basename(tau_tags["filename"])
        if fname in fname_to_idx:
            index = fname_to_idx[fname]
            for key in tau_additional_keys:
                raw_data[key][index] = tau_tags[key]

    # Sanity checks
    assert all(
        all(value is not None for value in raw_data[key]) for key in tau_additional_keys
    )
    assert all(len(values) == dataset_size for values in raw_data.values())

    if verbose >= 1:
        msg = f"Dataset {MACSCard.PRETTY_NAME} ({subset}) has been loaded. {len(next(iter(raw_data.values())))=})"
        logger.info(msg)

    return raw_data, annotator_id_to_competence


def download_macs_dataset(
    # Common args
    root: Union[str, Path, None] = None,
    subset: MACSSubset = MACSCard.DEFAULT_SUBSET,
    force: bool = False,
    verbose: int = 0,
    verify_files: bool = True,
    *,
    # MACS-specific args
    clean_archives: bool = True,
) -> None:
    """Prepare MACS data.

    :param root: Dataset root directory.
        defaults to ".".
    :param subset: The subset of MACS to use. Can be one of :attr:`~MACSCard.SUBSETS`.
        defaults to "full".
    :param force: If True, force to download again all files.
        defaults to False.
    :param verbose: Verbose level.
        defaults to 0.
    :param verify_files: If True, check all file already downloaded are valid.
        defaults to False.

    :param clean_archives: If True, remove the compressed archives from disk to save space.
        defaults to True.
    """

    root = _get_root(root)
    if not osp.isdir(root):
        raise RuntimeError(f"Cannot find root directory '{root}'.")

    macs_dpath = _get_macs_root(root)
    archives_dpath = _get_archives_dpath(root)
    audio_dpath = _get_audio_dpath(root)
    tau_meta_dpath = _get_tau_meta_dpath(root)

    for dpath in (archives_dpath, audio_dpath, tau_meta_dpath):
        os.makedirs(dpath, exist_ok=True)

    # Download MACS specific files
    for file_info in MACS_FILES.values():
        fname = file_info["fname"]
        fpath = osp.join(macs_dpath, fname)

        if not osp.isfile(fpath) or force:
            if verbose >= 1:
                logger.info(f"Downloading captions file '{fname}'...")

            url = file_info["url"]
            download_file(url, fpath, verbose=verbose)

        if verify_files:
            hash_value = file_info["hash_value"]
            file_hash_value = hash_file(fpath, hash_type="md5")
            if file_hash_value != hash_value:
                raise RuntimeError(
                    f"Invalid checksum for file '{fname}'. (expected md5 checksum '{hash_value}' but found '{file_hash_value}')\n"
                    f"Please try to remove manually the file '{fpath}' and rerun {MACSCard.PRETTY_NAME} download."
                )
            elif verbose >= 2:
                logger.debug(f"File '{fname}' has a valid checksum.")

    captions_fpath = osp.join(macs_dpath, MACS_FILES["captions"]["fname"])
    with open(captions_fpath, "r") as file:
        captions_data = yaml.safe_load(file)
    captions_data = captions_data["files"]

    # Download TAU Urban Sound audio archives files
    for i, file_info in enumerate(MACS_ARCHIVES_FILES.values()):
        zip_fname = file_info["fname"]
        zip_fpath = osp.join(archives_dpath, zip_fname)

        if not osp.isfile(zip_fpath) or force:
            if verbose >= 1:
                msg = f"Downloading audio zip file '{zip_fpath}'... ({i + 1}/{len(MACS_ARCHIVES_FILES)})"
                logger.info(msg)

            url = file_info["url"]
            download_file(url, zip_fpath, verbose=verbose)

        if verify_files:
            hash_value = file_info["hash_value"]
            file_hash_value = hash_file(zip_fpath, hash_type="md5")
            if file_hash_value != hash_value:
                raise RuntimeError(
                    f"Invalid checksum for file '{zip_fname}'. (expected md5 checksum '{hash_value}' but found '{file_hash_value}')\n"
                    f"Please try to remove manually the file '{zip_fpath}' and rerun {MACSCard.PRETTY_NAME} download."
                )
            elif verbose >= 2:
                logger.debug(f"File '{zip_fname}' has a valid checksum.")

    # Extract files from TAU Urban Sound archives
    macs_fnames = dict.fromkeys(data["filename"] for data in captions_data)
    for i, (name, file_info) in enumerate(MACS_ARCHIVES_FILES.items()):
        zip_fname = file_info["fname"]
        zip_fpath = osp.join(archives_dpath, zip_fname)

        if verbose >= 2:
            msg = f"Check to extract TAU Urban acoustic scenes archive {zip_fname=}..."
            logger.debug(msg)

        is_audio_archive = name.startswith("audio")
        if is_audio_archive:
            target_dpath = audio_dpath
        else:
            target_dpath = tau_meta_dpath

        with zipfile.ZipFile(zip_fpath, "r") as file:
            members_to_extract = [
                member
                for member in file.namelist()
                # Extract member if file if in captions yaml file and if the audio file is not already downloaded
                if (
                    (osp.basename(member) in macs_fnames or not is_audio_archive)
                    and not osp.isfile(osp.join(target_dpath, osp.basename(member)))
                )
            ]

            if verbose >= 1:
                msg = f"Extracting {len(members_to_extract)}/{len(file.namelist())} audio files from ZIP file '{zip_fname}'... ({i + 1}/{len(MACS_ARCHIVES_FILES)})"
                logger.info(msg)

            if len(members_to_extract) > 0:
                file.extractall(archives_dpath, members_to_extract)
                for member in members_to_extract:
                    extracted_fpath = osp.join(archives_dpath, member)
                    target_fpath = osp.join(target_dpath, osp.basename(member))
                    shutil.move(extracted_fpath, target_fpath)

    if clean_archives:
        if verbose >= 1:
            logger.info(f"Removing archives files in {archives_dpath}...")
        shutil.rmtree(archives_dpath, ignore_errors=True)

    audio_fnames = [name for name in os.listdir(audio_dpath) if name.endswith(".wav")]
    assert len(audio_fnames) == len(macs_fnames)

    if verbose >= 2:
        logger.debug(f"Dataset {MACSCard.PRETTY_NAME} ({subset}) has been prepared.")


def download_macs_datasets(
    # Common args
    root: Union[str, Path, None] = None,
    subsets: Union[MACSSubset, Iterable[MACSSubset]] = MACSCard.DEFAULT_SUBSET,
    force: bool = False,
    verbose: int = 0,
    *,
    # MACS-specific args
    clean_archives: bool = True,
    verify_files: bool = True,
) -> None:
    """Function helper to download a list of subsets. See :func:`~aac_datasets.datasets.functional.macs.download_macs_dataset` for details."""
    if isinstance(subsets, str):
        subsets = [subsets]
    else:
        subsets = list(subsets)

    kwargs: Dict[str, Any] = dict(
        root=root,
        force=force,
        verbose=verbose,
        clean_archives=clean_archives,
        verify_files=verify_files,
    )
    for subset in subsets:
        download_macs_dataset(
            subset=subset,
            **kwargs,
        )


def _get_macs_root(root: str) -> str:
    return osp.join(root, "MACS")


def _get_archives_dpath(root: str) -> str:
    return osp.join(_get_macs_root(root), "archives")


def _get_audio_dpath(root: str) -> str:
    return osp.join(_get_macs_root(root), "audio")


def _get_tau_meta_dpath(root: str) -> str:
    return osp.join(_get_macs_root(root), "tau_meta")


def _is_prepared_macs(root: str) -> bool:
    audio_dpath = _get_audio_dpath(root)
    if not osp.isdir(audio_dpath):
        return False
    captions_fpath = osp.join(_get_macs_root(root), MACS_FILES["captions"]["fname"])
    if not osp.isfile(captions_fpath):
        return False

    with open(captions_fpath, "r") as file:
        data = yaml.safe_load(file)
    data = data["files"]
    fnames = os.listdir(audio_dpath)
    return len(data) == len(fnames)


# Internal typing to make easier to add new links without error
MACSLinkType = Literal["licence", "captions", "annotators_competences"]

# MACS-specific files links.
MACS_FILES: Dict[MACSLinkType, LinkInfoHash] = {
    "licence": {
        "fname": "LICENSE.txt",
        "url": "https://zenodo.org/record/5114771/files/LICENSE.txt?download=1",
        "hash_value": "d3086f4517cccc32c1bb3a081b07cfa1",
    },
    "captions": {
        "fname": "MACS.yaml",
        "url": "https://zenodo.org/record/5114771/files/MACS.yaml?download=1",
        "hash_value": "23fcb2ebd0b109094034ef9e87972256",
    },
    "annotators_competences": {
        "fname": "MACS_competence.csv",
        "url": "https://zenodo.org/record/5114771/files/MACS_competence.csv?download=1",
        "hash_value": "4dfe9f951f0af9f29cb7952ec030370a",
    },
}

# TAU_URBAN_ACOUSTIC archives files links.
TAU_URBAN_ACOUSTIC_DEV_FILES = {
    "audio.1": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.1.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.1.zip?download=1",
        "hash_value": "aca4ebfd9ed03d5f747d6ba8c24bc728",
    },
    "audio.10": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.10.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.10.zip?download=1",
        "hash_value": "0ffbf60006da520cc761fb74c878b98b",
    },
    "audio.11": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.11.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.11.zip?download=1",
        "hash_value": "599055d93b4c11057c29be2df54538d4",
    },
    "audio.12": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.12.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.12.zip?download=1",
        "hash_value": "98b8d162ff3665695c4c910e6c372cc8",
    },
    "audio.13": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.13.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.13.zip?download=1",
        "hash_value": "a356c08b1a5a21d433eba37ef87587f4",
    },
    "audio.14": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.14.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.14.zip?download=1",
        "hash_value": "f8969771e7faf7dd471d1cf78b0cf011",
    },
    "audio.15": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.15.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.15.zip?download=1",
        "hash_value": "4758c4b0fb7484faa632266e78850820",
    },
    "audio.16": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.16.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.16.zip?download=1",
        "hash_value": "a18acad9ede8ea76574216feb887f0bc",
    },
    "audio.17": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.17.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.17.zip?download=1",
        "hash_value": "1af7703484632f340da5c33662dc9632",
    },
    "audio.18": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.18.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.18.zip?download=1",
        "hash_value": "b67402bf3e08f4da394a7c18756c0fd2",
    },
    "audio.19": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.19.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.19.zip?download=1",
        "hash_value": "035db315f19106eb848b6f9b32bcc47c",
    },
    "audio.2": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.2.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.2.zip?download=1",
        "hash_value": "c4f170408ce77c8c70c532bf268d7be0",
    },
    "audio.20": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.20.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.20.zip?download=1",
        "hash_value": "9cb28c74911bf8a3eadcf53f50a5b5d6",
    },
    "audio.21": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.21.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.21.zip?download=1",
        "hash_value": "0e44ed85c88ec036a9725b4dd1dfaea0",
    },
    "audio.3": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.3.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.3.zip?download=1",
        "hash_value": "c7214a07211f10f3250290d05e72c37e",
    },
    "audio.4": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.4.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.4.zip?download=1",
        "hash_value": "a6a62110f6699cf4432072acb1dffda6",
    },
    "audio.5": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.5.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.5.zip?download=1",
        "hash_value": "091a0b6d3c84b8e60e46940aa7d4a8a0",
    },
    "audio.6": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.6.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.6.zip?download=1",
        "hash_value": "114f4ca13e074391b98a1cfd8140de65",
    },
    "audio.7": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.7.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.7.zip?download=1",
        "hash_value": "5951dd2968f7a514e2afbe279c4f060d",
    },
    "audio.8": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.8.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.8.zip?download=1",
        "hash_value": "b0b63dc95b327e1509857c8d8a663cc3",
    },
    "audio.9": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.9.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.9.zip?download=1",
        "hash_value": "3c32a693a6b111ffb957be3c1dd22e9b",
    },
    "doc": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.doc.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.doc.zip?download=1",
        "hash_value": "1f6879544e80da70099a191613e7e51f",
    },
    "meta": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.meta.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.meta.zip?download=1",
        "hash_value": "09782f2097e4735687af73c44919329c",
    },
}

# List of TAU_URBAN_ACOUSTIC archives containing at least 1 MACS audio file.
MACS_ARCHIVES_FILES = {
    name: TAU_URBAN_ACOUSTIC_DEV_FILES[name]
    for name in (
        "audio.1",
        "audio.10",
        "audio.11",
        "audio.12",
        "audio.13",
        "audio.2",
        "audio.3",
        "audio.9",
        "meta",
    )
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Functional interface to load/download datasets.

Each module contains the its own logic for data preparation, but returns a similar dict structure for each dataset class.
"""

from .audiocaps import AudioCapsCard, download_audiocaps_dataset, load_audiocaps_dataset
from .clotho import ClothoCard, download_clotho_dataset, load_clotho_dataset
from .macs import MACSCard, download_macs_dataset, load_macs_dataset
from .wavcaps import WavCapsCard, download_wavcaps_dataset, load_wavcaps_dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    # To support torchaudio >= 2.1.0
    from torchaudio import AudioMetaData  # type: ignore
except ImportError:
    from torchaudio.backend.common import AudioMetaData


__all__ = ["AudioMetaData"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os.path as osp
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    List,
    Literal,
    Mapping,
    Optional,
    Union,
    overload,
)

from torch import Tensor
from typing_extensions import NotRequired, TypedDict, TypeVar

from aac_datasets.datasets.base import AACDataset
from aac_datasets.datasets.functional.clotho import (
    ClothoCard,
    ClothoSubset,
    ClothoVersion,
    _get_audio_subset_dpath,
    download_clotho_dataset,
    load_clotho_dataset,
)
from aac_datasets.utils.globals import _get_root

logger = logging.getLogger(__name__)


class ClothoDevValEvalItem(TypedDict):
    r"""Class representing a single Clotho item."""

    # Common attributes
    audio: Tensor
    captions: List[str]
    dataset: str
    fname: str
    index: int
    subset: ClothoSubset
    sr: int
    duration: float
    # Clotho-specific attributes
    keywords: List[str]
    sound_id: str  # warning: some files contains "Not found"
    sound_link: str  # warning: some files contains "NA"
    start_end_samples: str  # warning: some files contains ""
    manufacturer: str
    license: str


class ClothoDCASEAACTestItem(TypedDict):
    r"""Class representing a single Clotho item."""

    # Common attributes
    audio: Tensor
    dataset: str
    fname: str
    index: int
    subset: ClothoSubset
    sr: int
    duration: float
    # Clotho-specific attributes
    start_end_samples: str  # warning: some files contains ""
    manufacturer: str
    license: str


class ClothoDCASEAACAnalysisItem(TypedDict):
    r"""Class representing a single Clotho item."""

    # Common attributes
    audio: Tensor
    dataset: str
    fname: str
    index: int
    subset: ClothoSubset
    sr: int
    duration: float


class ClothoDCASET2AAudioItem(TypedDict):
    r"""Class representing a single Clotho item."""

    # Common attributes
    audio: Tensor
    dataset: str
    fname: str
    index: int
    subset: ClothoSubset
    sr: int
    duration: float
    # Clotho-specific attributes
    keywords: List[str]
    sound_id: str  # warning: some files contains "Not found"
    sound_link: str  # warning: some files contains "NA"
    start_end_samples: str  # warning: some files contains ""
    manufacturer: str
    license: str


class ClothoDCASET2ACaptionsItem(TypedDict):
    r"""Class representing a single Clotho item."""

    # Common attributes
    captions: List[str]
    dataset: str
    index: int
    subset: ClothoSubset


class ClothoItem(TypedDict):
    r"""Class representing a single Clotho item."""

    # Common attributes
    audio: NotRequired[Tensor]
    captions: NotRequired[List[str]]
    dataset: str
    fname: NotRequired[str]
    index: int
    subset: ClothoSubset
    sr: NotRequired[int]
    duration: NotRequired[float]
    # Clotho-specific attributes
    keywords: NotRequired[List[str]]
    sound_id: NotRequired[str]  # warning: some files contains "Not found"
    sound_link: NotRequired[str]  # warning: some files contains "NA"
    start_end_samples: NotRequired[str]  # warning: some files contains ""
    manufacturer: NotRequired[str]
    license: NotRequired[str]


T_ClothoItem = TypeVar(
    "T_ClothoItem", bound=Mapping, covariant=True, default=ClothoItem
)


class Clotho(Generic[T_ClothoItem], AACDataset[T_ClothoItem]):
    r"""Unofficial Clotho PyTorch dataset.

    Subsets available are 'train', 'val', 'eval', 'dcase_aac_test', 'dcase_aac_analysis', 'dcase_t2a_audio' and 'dcase_t2a_captions'.

    Audio are waveform sounds of 15 to 30 seconds, sampled at 44100 Hz.
    Target is a list of 5 different sentences strings describing an audio sample.
    The maximal number of words in captions is 20.

    Clotho V1 Paper: https://arxiv.org/pdf/1910.09387.pdf

    .. code-block:: text
        :caption:  Dataset folder tree for version 'v2.1', with all subsets

        {root}
        └── CLOTHO_v2.1
            ├── archives
            |   └── (5 7z files, ~8.9GB)
            ├── clotho_audio_files
            │   ├── clotho_analysis
            │   │    └── (8360 wav files, ~19GB)
            │   ├── development
            │   │    └── (3839 wav files, ~7.1GB)
            │   ├── evaluation
            │   │    └── (1045 wav files, ~2.0GB)
            │   ├── test
            │   |    └── (1043 wav files, ~2.0GB)
            │   ├── test_retrieval_audio
            │   |    └── (1000 wav files, ~2.0GB)
            │   └── validation
            │        └── (1045 wav files, ~2.0GB)
            └── clotho_csv_files
                ├── clotho_captions_development.csv
                ├── clotho_captions_evaluation.csv
                ├── clotho_captions_validation.csv
                ├── clotho_metadata_development.csv
                ├── clotho_metadata_evaluation.csv
                ├── clotho_metadata_test.csv
                ├── clotho_metadata_validation.csv
                ├── retrieval_audio_metadata.csv
                └── retrieval_captions.csv
    """

    # Common globals
    CARD: ClassVar[ClothoCard] = ClothoCard()

    # Clotho-specific globals
    INVALID_SOUND_ID: ClassVar[str] = "Not found"
    INVALID_SOUND_LINK: ClassVar[str] = "NA"
    INVALID_START_END_SAMPLES: ClassVar[str] = ""

    # Initialization
    @overload
    def __init__(
        self: "Clotho[ClothoDevValEvalItem]",
        # Common args
        root: Union[str, Path, None] = None,
        subset: Literal["dev", "val", "eval"] = "dev",
        download: bool = False,
        transform: Optional[Callable[[ClothoItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        *,
        # Clotho-specific args
        clean_archives: bool = True,
        flat_captions: bool = False,
        version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
    ) -> None: ...

    @overload
    def __init__(
        self: "Clotho[ClothoDCASEAACTestItem]",
        # Common args
        root: Union[str, Path, None] = None,
        *,
        subset: Literal["dcase_aac_test"],
        download: bool = False,
        transform: Optional[Callable[[ClothoItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        # Clotho-specific args
        clean_archives: bool = True,
        flat_captions: bool = False,
        version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
    ) -> None: ...

    @overload
    def __init__(
        self: "Clotho[ClothoDCASEAACAnalysisItem]",
        # Common args
        root: Union[str, Path, None] = None,
        *,
        subset: Literal["dcase_aac_analysis"],
        download: bool = False,
        transform: Optional[Callable[[ClothoItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        # Clotho-specific args
        clean_archives: bool = True,
        flat_captions: bool = False,
        version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
    ) -> None: ...

    @overload
    def __init__(
        self: "Clotho[ClothoDCASET2AAudioItem]",
        # Common args
        root: Union[str, Path, None] = None,
        *,
        subset: Literal["dcase_t2a_audio"],
        download: bool = False,
        transform: Optional[Callable[[ClothoItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        # Clotho-specific args
        clean_archives: bool = True,
        flat_captions: bool = False,
        version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
    ) -> None: ...

    @overload
    def __init__(
        self: "Clotho[ClothoDCASET2ACaptionsItem]",
        # Common args
        root: Union[str, Path, None] = None,
        *,
        subset: Literal["dcase_t2a_captions"],
        download: bool = False,
        transform: Optional[Callable[[ClothoItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        # Clotho-specific args
        clean_archives: bool = True,
        flat_captions: bool = False,
        version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
    ) -> None: ...

    def __init__(
        self: "Clotho[ClothoItem]",
        # Common args
        root: Union[str, Path, None] = None,
        subset: ClothoSubset = ClothoCard.DEFAULT_SUBSET,
        download: bool = False,
        transform: Optional[Callable[[ClothoItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        *,
        # Clotho-specific args
        clean_archives: bool = True,
        flat_captions: bool = False,
        version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
    ) -> None:
        """
        :param root: The parent of the dataset root directory.
            Note: The data is stored in the 'CLOTHO_{version}' subdirectory.
            defaults to ".".
        :param subset: The subset of Clotho to use. Can be one of :attr:`~ClothoCard.SUBSETS`.
            defaults to "dev".
        :param download: Download the dataset if download=True and if the dataset is not already downloaded.
            defaults to False.
        :param transform: The transform to apply to the global dict item. This transform is applied only in getitem method when argument is an integer.
            defaults to None.
        :param verbose: Verbose level to use. Can be 0 or 1.
            defaults to 0.
        :param force_download: If True, force to re-download file even if they exists on disk.
            defaults to False.
        :param verify_files: If True, check hash value when possible.
            defaults to False.

        :param clean_archives: If True, remove the compressed archives from disk to save space.
            defaults to True.
        :param flat_captions: If True, map captions to audio instead of audio to caption.
            defaults to True.
        :param version: The version of the dataset. Can be one of :attr:`~ClothoCard.VERSIONS`.
            defaults to 'v2.1'.
        """
        if version not in ClothoCard.VERSIONS:
            msg = f"Invalid Clotho argument {version=}. Must be one of {ClothoCard.VERSIONS}."
            raise ValueError(msg)

        if version == "v2":
            msg = (
                f"The version '{version}' of the Clotho dataset contains minor some errors in file names and few corrupted files."
                f"Please consider using the fixed version 'v2.1'."
            )
            logger.warning(msg)

        if subset not in ClothoCard.SUBSETS:
            msg = f"Invalid Clotho argument subset={subset} for {version=}. Must be one of {ClothoCard.SUBSETS}."
            raise ValueError(msg)

        root = _get_root(root)

        if download:
            download_clotho_dataset(
                root=root,
                subset=subset,
                force=force_download,
                verbose=verbose,
                verify_files=verify_files,
                clean_archives=clean_archives,
                version=version,
            )

        # Exclude some columns containing empty values for several subsets
        column_names = list(ClothoItem.__required_keys__) + list(  # type: ignore
            ClothoItem.__optional_keys__  # type: ignore
        )
        if subset == "dcase_aac_test":
            removed_columns = ("captions", "sound_id", "keywords", "sound_link")
        elif subset == "dcase_aac_analysis":
            removed_columns = (
                "captions",
                "sound_id",
                "keywords",
                "sound_link",
                "license",
                "manufacturer",
                "start_end_samples",
            )
        elif subset == "dcase_t2a_audio":
            removed_columns = ("captions",)
        elif subset == "dcase_t2a_captions":
            removed_columns = (
                "audio",
                "sr",
                "duration",
                "fname",
                "sound_id",
                "keywords",
                "sound_link",
                "license",
                "manufacturer",
                "start_end_samples",
            )
        else:
            removed_columns = ()

        for name in removed_columns:
            column_names.remove(name)

        raw_data = load_clotho_dataset(
            root=root,
            subset=subset,
            verbose=verbose,
            version=version,
        )

        size = len(next(iter(raw_data.values())))
        raw_data["dataset"] = [ClothoCard.NAME] * size
        raw_data["subset"] = [subset] * size
        raw_data["index"] = list(range(size))

        if "audio" not in removed_columns:
            audio_subset_dpath = _get_audio_subset_dpath(root, version, subset)
            assert audio_subset_dpath is not None, (
                "Internal error. (expected audio column but audio dname is None)"
            )
            raw_data["fpath"] = [
                osp.join(audio_subset_dpath, fname) for fname in raw_data["fname"]
            ]

        super().__init__(
            raw_data=raw_data,
            transform=transform,
            column_names=column_names,
            flat_captions=flat_captions,
            sr=ClothoCard.SAMPLE_RATE,
            verbose=verbose,
        )
        self._root = root
        self._subset: ClothoSubset = subset
        self._download = download
        self._version: ClothoVersion = version

        if "audio" not in removed_columns:
            self.add_online_columns(
                {
                    "audio": Clotho._load_audio,
                    "audio_metadata": Clotho._load_audio_metadata,
                    "duration": Clotho._load_duration,
                    "num_channels": Clotho._load_num_channels,
                    "num_frames": Clotho._load_num_frames,
                    "sr": Clotho._load_sr,
                }
            )

    # Properties
    @property
    def download(self) -> bool:
        return self._download

    @property
    def root(self) -> str:
        return self._root

    @property
    def sr(self) -> int:
        return self._sr  # type: ignore

    @property
    def subset(self) -> ClothoSubset:
        return self._subset

    @property
    def version(self) -> ClothoVersion:
        return self._version

    # Magic methods
    def __repr__(self) -> str:
        repr_dic = {
            "subset": self._subset,
            "size": len(self),
            "num_columns": len(self.column_names),
            "version": self._version,
        }
        repr_str = ", ".join(f"{k}={v}" for k, v in repr_dic.items())
        return f"{ClothoCard.PRETTY_NAME}({repr_str})"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import csv
import logging
import os
import os.path as osp
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, get_args
from zipfile import ZipFile

import pythonwrench as pw
from py7zr import SevenZipFile
from torchwrench.hub.download import download_file, hash_file
from typing_extensions import Literal

from aac_datasets.datasets.functional.common import DatasetCard, LinkInfoHash
from aac_datasets.utils.globals import _get_root

logger = logging.getLogger(__name__)


ClothoSubset = Literal[
    "dev",
    "val",
    "eval",
    "dcase_aac_test",
    "dcase_aac_analysis",
    "dcase_t2a_audio",
    "dcase_t2a_captions",
]
ClothoVersion = Literal["v1", "v2", "v2.1"]


class ClothoCard(DatasetCard):
    ANNOTATIONS_CREATORS: Tuple[str, ...] = ("crowdsourced",)
    CAPTIONS_PER_AUDIO: Dict[ClothoSubset, int] = {
        "dev": 5,
        "val": 5,
        "eval": 5,
        "dcase_aac_test": 0,
        "dcase_aac_analysis": 0,
        "dcase_t2a_audio": 0,
        "dcase_t2a_captions": 1,
    }
    CITATION: str = r"""
    @inproceedings{Drossos_2020_icassp,
        title        = {Clotho: an Audio Captioning Dataset},
        author       = {Drossos, Konstantinos and Lipping, Samuel and Virtanen, Tuomas},
        year         = 2020,
        booktitle    = {Proc. IEEE Int. Conf. Acoustic., Speech and Signal Process. (ICASSP)},
        pages        = {736--740},
        abstract     = {Audio captioning is the novel task of general audio content description using free text. It is an intermodal translation task (not speech-to-text), where a system accepts as an input an audio signal and outputs the textual description (i.e. the caption) of that signal. In this paper we present Clotho, a dataset for audio captioning consisting of 4981 audio samples of 15 to 30 seconds duration and 24 905 captions of eight to 20 words length, and a baseline method to provide initial results. Clotho is built with focus on audio content and caption diversity, and the splits of the data are not hampering the training or evaluation of methods. All sounds are from the Freesound platform, and captions are crowdsourced using Amazon Mechanical Turk and annotators from English speaking countries. Unique words, named entities, and speech transcription are removed with post-processing. Clotho is freely available online (https://zenodo.org/record/3490684).}
    }
    """
    DEFAULT_SUBSET: ClothoSubset = "dev"
    DEFAULT_VERSION: ClothoVersion = "v2.1"
    HOMEPAGE: str = "https://zenodo.org/record/3490684"
    LANGUAGE: Tuple[str, ...] = ("en",)
    LANGUAGE_DETAILS: Tuple[str, ...] = ("en-US",)
    NAME: str = "clotho"
    N_CHANNELS: int = 1
    PRETTY_NAME: str = "Clotho"
    SAMPLE_RATE: int = 44_100  # Hz
    SIZE_CATEGORIES: Tuple[str, ...] = ("1K<n<10K",)
    SUBSETS: Tuple[ClothoSubset, ...] = get_args(ClothoSubset)
    TASK_CATEGORIES: Tuple[str, ...] = ("audio-to-text", "text-to-audio")
    VERSIONS: Tuple[ClothoVersion, ...] = get_args(ClothoVersion)


def load_clotho_dataset(
    # Common args
    root: Union[str, Path, None] = None,
    subset: ClothoSubset = ClothoCard.DEFAULT_SUBSET,
    verbose: int = 0,
    *,
    # Clotho-specific args
    version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
) -> Dict[str, List[Any]]:
    """Load Clotho metadata.

    :param root: Dataset root directory.
        defaults to ".".
    :param subset: The subset of Clotho to use. Can be one of :attr:`~ClothoCard.SUBSETS`.
        defaults to "dev".
    :param verbose: Verbose level.
        defaults to 0.

    :param version: The version of the dataset. Can be one of :attr:`~ClothoCard.VERSIONS`.
        defaults to 'v2.1'.
    :returns: A dictionnary of lists containing each metadata.
    """
    root = _get_root(root)
    if not _is_prepared_clotho(root, version, subset):
        msg = f"Cannot load data: clotho_{subset} is not prepared in data {root=}. Please use download=True in dataset constructor."
        raise RuntimeError(msg)

    # Read fpath of .wav audio files
    links = _CLOTHO_LINKS[version][subset]
    csv_dpath = _get_csv_dpath(root, version)

    # Read Clotho files
    if "captions" in links.keys():
        captions_fname = links["captions"]["fname"]
        captions_fpath = osp.join(csv_dpath, captions_fname)

        # Keys: file_name, caption_1, caption_2, caption_3, caption_4, caption_5
        with open(captions_fpath, "r") as file:
            reader = csv.DictReader(file)
            captions_data = list(reader)

        if subset == "dcase_t2a_captions":
            captions_data = [
                pw.union_dicts([data, {"file_name": f"no_fname_{i}"}])
                for i, data in enumerate(captions_data)
            ]

    else:
        captions_data = []

    if "metadata" in links.keys():
        metadata_fname = links["metadata"]["fname"]
        metadata_fpath = osp.join(csv_dpath, metadata_fname)

        # Keys: file_name, keywords, sound_id, sound_link, start_end_samples, manufacturer, license
        if version in ("v2", "v2.1"):
            encoding = "ISO-8859-1"
        else:
            encoding = None

        with open(metadata_fpath, "r", encoding=encoding) as file:
            delimiter = ";" if subset == "dcase_aac_test" else ","
            reader = csv.DictReader(file, delimiter=delimiter)
            metadata = list(reader)
    else:
        metadata = []

    if "captions" in links.keys():
        # note: "dev", "val", "eval"
        fnames_lst = [line["file_name"] for line in captions_data]
    elif "metadata" in links.keys():
        # note: for "dcase_aac_test" subset which do not have captions CSV file
        fnames_lst = [line["file_name"] for line in metadata]
    else:
        # note 1: for "dcase_aac_analysis" subset which do not have any CSV file
        # note 2: force sorted list to have the same order on all OS
        audio_subset_dpath = _get_audio_subset_dpath(root, version, subset)
        if audio_subset_dpath is None:
            msg = f"INTERNAL ERROR: Invalid audio subset dirpath. (found audio_subset_dpath={audio_subset_dpath}, with {subset=})"
            raise RuntimeError(msg)
        fnames_lst = list(sorted(os.listdir(audio_subset_dpath)))

    idx_to_fname = {i: fname for i, fname in enumerate(fnames_lst)}
    fname_to_idx = {fname: i for i, fname in idx_to_fname.items()}
    dataset_size = len(fnames_lst)

    # Process each item field
    if len(metadata) > 0:
        subset_metadata_keys = [key for key in _METADATA_KEYS if key in metadata[0]]
    else:
        subset_metadata_keys = []

    all_captions_lst = [[] for _ in range(dataset_size)]

    if subset != "dcase_t2a_captions":
        captions_keys = _CAPTIONS_KEYS
    else:
        captions_keys = ("caption",)

    for line in captions_data:
        fname = line["file_name"]
        index = fname_to_idx[fname]
        all_captions_lst[index] = [line[caption_key] for caption_key in captions_keys]

    all_metadata_dic: Dict[str, List[Any]] = {
        key: [None for _ in range(dataset_size)] for key in subset_metadata_keys
    }
    for line in metadata:
        fname = line["file_name"]
        if fname not in fname_to_idx:
            msg = f"Cannot find metadata fname={fname} in captions file. {subset=})"
            raise KeyError(msg)
        index = fname_to_idx[fname]
        for key in subset_metadata_keys:
            # The test subset does not have keywords in metadata, but has sound_id, sound_link, etc.
            if key in line:
                all_metadata_dic[key][index] = line[key]

    raw_data = {
        "fname": fnames_lst,
        "captions": all_captions_lst,
    }
    raw_data.update(all_metadata_dic)

    if "keywords" in raw_data:
        # Split keywords into List[str]
        raw_data["keywords"] = [
            keywords.split(";") if keywords is not None else []
            for keywords in raw_data["keywords"]
        ]

    if subset == "dcase_t2a_audio":
        # Temporary patch to avoid file loading errors
        # indexes: 53, 521, 677
        replaces = {
            "raindrops on metal: police background.wav": "raindrops on metal_police background.wav",
            "Intersection Wet : Metro Pass.wav": "Intersection Wet_Metro Pass.wav",
            "Kitchen Roomtone w: Dripping Faucet_1-2.wav": "Kitchen Roomtone w_Dripping Faucet_1-2.wav",
        }
        raw_data["fname"] = [replaces.get(fname, fname) for fname in raw_data["fname"]]

    if verbose >= 1:
        msg = f"Dataset {ClothoCard.PRETTY_NAME} ({subset}) has been loaded. {len(next(iter(raw_data.values())))=})"
        logger.info(msg)
    return raw_data


def download_clotho_dataset(
    # Common args
    root: Union[str, Path, None] = None,
    subset: ClothoSubset = ClothoCard.DEFAULT_SUBSET,
    force: bool = False,
    verbose: int = 0,
    verify_files: bool = True,
    *,
    # Clotho-specific args
    clean_archives: bool = True,
    version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
) -> None:
    """Prepare Clotho data.

    :param root: Dataset root directory.
        defaults to ".".
    :param subset: The subset of Clotho to use. Can be one of :attr:`~ClothoCard.SUBSETS`.
        defaults to "dev".
    :param force: If True, force to download again all files.
        defaults to False.
    :param verbose: Verbose level.
        defaults to 0.
    :param verify_files: If True, check all file already downloaded are valid.
        defaults to False.

    :param clean_archives: If True, remove the compressed archives from disk to save space.
        defaults to True.
    :param version: The version of the dataset. Can be one of :attr:`~ClothoCard.VERSIONS`.
        defaults to 'v2.1'.
    """
    if subset == "val" and version == "v1":
        logger.error(
            f"Clotho version '{version}' does not have '{subset}' subset. It will be ignored."
        )
        return None

    root = _get_root(root)
    if not osp.isdir(root):
        raise RuntimeError(f"Cannot find root directory '{root}'.")

    archives_dpath = _get_archives_dpath(root, version)
    audio_dpath = _get_audio_dpath(root, version)
    csv_dpath = _get_csv_dpath(root, version)

    for dpath in (archives_dpath, audio_dpath, csv_dpath):
        os.makedirs(dpath, exist_ok=True)

    if verbose >= 1:
        logger.info(f"Start to download files for clotho_{subset}...")

    links = copy.deepcopy(_CLOTHO_LINKS[version][subset])
    EXTENSIONS = ("7z", "csv", "zip")

    # Download csv and 7z files
    for file_info in links.values():
        fname, url, hash_value = (
            file_info["fname"],
            file_info["url"],
            file_info["hash_value"],
        )
        extension = fname.split(".")[-1]

        if extension in ("7z", "zip"):
            dpath = archives_dpath
        elif extension == "csv":
            dpath = csv_dpath
        else:
            msg = f"Found invalid {extension=}. Must be one of {EXTENSIONS}."
            raise RuntimeError(msg)

        fpath = osp.join(dpath, fname)
        if not osp.isfile(fpath) or force:
            if verbose >= 1:
                logger.info(f"Download and check file '{fname}' from {url=}...")

            download_file(url, fpath, verbose=verbose)

        elif verbose >= 1:
            logger.info(f"File '{fname}' is already downloaded.")

        if verify_files:
            hash_value = file_info["hash_value"]
            file_hash_value = hash_file(fpath, hash_type="md5")
            if file_hash_value != hash_value:
                raise RuntimeError(
                    f"Invalid checksum for file '{fname}'. (expected md5 checksum '{hash_value}' but found '{file_hash_value}')\n"
                    f"Please try to remove manually the file '{fpath}' and rerun {ClothoCard.PRETTY_NAME} download."
                )
            elif verbose >= 2:
                logger.debug(f"File '{fname}' has a valid checksum.")

    # Extract audio files from archives
    audio_subset_dpath = _get_audio_subset_dpath(root, version, subset)
    if audio_subset_dpath is not None:
        for file_info in links.values():
            fname = file_info["fname"]
            extension = fname.split(".")[-1]

            if extension == "csv":
                continue

            if extension not in ("7z", "zip"):
                logger.error(
                    f"Found unexpected {extension=} for downloaded file '{fname}'. Expected one of {EXTENSIONS}."
                )
                continue

            fpath = osp.join(archives_dpath, fname)

            if verbose >= 1:
                logger.info(f"Extract archive file {fname=}...")

            if extension == "7z":
                archive_file = SevenZipFile(fpath)
                compressed_fnames = [
                    osp.basename(fname) for fname in archive_file.getnames()
                ]
            elif extension == "zip":
                archive_file = ZipFile(fpath)
                compressed_fnames = [
                    osp.basename(file.filename) for file in archive_file.filelist
                ]
            else:
                raise RuntimeError(f"Invalid extension '{extension}'.")

            # Ignore dir name from archive file
            compressed_fnames = [
                fname for fname in compressed_fnames if fname.endswith(".wav")
            ]
            extracted_fnames = (
                os.listdir(audio_subset_dpath) if osp.isdir(audio_subset_dpath) else []
            )

            if set(extracted_fnames) != set(compressed_fnames):
                # For dcase_t2a_audio subset, the name of the audio dname is also "test", so we need to move the audio files to another folder named "test_retrieval_audio".
                if subset == "dcase_t2a_audio":
                    target_dpath = audio_subset_dpath
                    os.makedirs(target_dpath, exist_ok=True)
                else:
                    target_dpath = audio_dpath

                archive_file.extractall(target_dpath)

                if subset == "dcase_t2a_audio":
                    extracted_dpath = osp.join(target_dpath, "test")
                    for fname in os.listdir(extracted_dpath):
                        os.rename(
                            osp.join(extracted_dpath, fname),
                            osp.join(target_dpath, fname),
                        )
                    os.rmdir(extracted_dpath)

                # Check if files is good now
                extracted_fnames = os.listdir(audio_subset_dpath)
                if set(extracted_fnames) != set(compressed_fnames):
                    found_but_not_expected = len(
                        set(extracted_fnames).difference(set(compressed_fnames))
                    )
                    expected_but_not_found = len(
                        set(compressed_fnames).difference(set(extracted_fnames))
                    )

                    raise RuntimeError(
                        f"Invalid number of audios extracted, found {len(extracted_fnames)} files but expected the same {len(compressed_fnames)} files. "
                        f"(with found_but_not_expected={found_but_not_expected} and {expected_but_not_found=})"
                    )

            archive_file.close()

    if clean_archives:
        for file_info in links.values():
            fname = file_info["fname"]
            extension = fname.split(".")[-1]
            if extension not in ("7z", "zip"):
                continue

            fpath = osp.join(archives_dpath, fname)
            if verbose >= 1:
                logger.info(f"Removing archive file {osp.basename(fpath)}...")
            os.remove(fpath)

    if verbose >= 2:
        logger.debug(f"Dataset {ClothoCard.PRETTY_NAME} ({subset}) has been prepared.")


def download_clotho_datasets(
    # Common args
    root: Union[str, Path, None] = None,
    subsets: Union[ClothoSubset, Iterable[ClothoSubset]] = ClothoCard.DEFAULT_SUBSET,
    force: bool = False,
    verbose: int = 0,
    *,
    # Clotho-specific args
    clean_archives: bool = True,
    verify_files: bool = True,
    version: ClothoVersion = ClothoCard.DEFAULT_VERSION,
) -> None:
    """Function helper to download a list of subsets. See :func:`~aac_datasets.datasets.functional.clotho.download_clotho_dataset` for details."""
    if isinstance(subsets, str):
        subsets = [subsets]
    else:
        subsets = list(subsets)

    kwargs: Dict[str, Any] = dict(
        root=root,
        force=force,
        verbose=verbose,
        clean_archives=clean_archives,
        verify_files=verify_files,
        version=version,
    )
    for subset in subsets:
        download_clotho_dataset(
            subset=subset,
            **kwargs,
        )


def _get_clotho_root(root: str, version: ClothoVersion) -> str:
    return osp.join(root, f"CLOTHO_{version}")


def _get_archives_dpath(root: str, version: ClothoVersion) -> str:
    return osp.join(_get_clotho_root(root, version), "archives")


def _get_audio_dpath(root: str, version: ClothoVersion) -> str:
    return osp.join(_get_clotho_root(root, version), "clotho_audio_files")


def _get_csv_dpath(root: str, version: ClothoVersion) -> str:
    return osp.join(_get_clotho_root(root, version), "clotho_csv_files")


def _get_audio_subset_dpath(
    root: str, version: ClothoVersion, subset: ClothoSubset
) -> Optional[str]:
    dname = _CLOTHO_AUDIO_DNAMES[subset]
    if dname is None:
        return None

    return osp.join(
        _get_clotho_root(root, version),
        "clotho_audio_files",
        dname,
    )


def _is_prepared_clotho(
    root: str,
    version: ClothoVersion,
    subset: ClothoSubset,
) -> bool:
    audio_dpath = _get_audio_dpath(root, version)
    csv_dpath = _get_csv_dpath(root, version)
    if not all(map(osp.isdir, (audio_dpath, csv_dpath))):
        return False

    links = _CLOTHO_LINKS[version][subset]

    if "captions" in links:
        captions_fname = links["captions"]["fname"]
        captions_fpath = osp.join(csv_dpath, captions_fname)

        if not osp.isfile(captions_fpath):
            return False

    if "metadata" in links:
        metadata_fname = links["metadata"]["fname"]
        metadata_fpath = osp.join(csv_dpath, metadata_fname)
        if not osp.isfile(metadata_fpath):
            return False

    if "audio_archive" in links:
        audio_subset_dpath = _get_audio_subset_dpath(root, version, subset)
        if audio_subset_dpath is None:
            msg = f"INTERNAL ERROR: Invalid audio subset dirpath. (found audio_subset_dpath={audio_subset_dpath}, with {subset=})"
            raise RuntimeError(msg)
        if not osp.isdir(audio_subset_dpath):
            return False

        audio_fnames = os.listdir(audio_subset_dpath)
        if "captions" in links:
            captions_fname = links["captions"]["fname"]
            captions_fpath = osp.join(csv_dpath, captions_fname)
            with open(captions_fpath, "r") as file:
                reader = csv.DictReader(file)
                lines = list(reader)
            return len(audio_fnames) == len(lines)
        else:
            return len(audio_fnames) > 0

    else:
        return True


# Audio directory names per subset
_CLOTHO_AUDIO_DNAMES: Dict[ClothoSubset, Optional[str]] = {
    "dev": "development",
    "val": "validation",
    "eval": "evaluation",
    "dcase_aac_test": "test",
    "dcase_aac_analysis": "clotho_analysis",
    "dcase_t2a_audio": "test_retrieval_audio",
    "dcase_t2a_captions": None,
}


# Internal typing to make easier to add new links without error
_ClothoLinkType = Literal["audio_archive", "captions", "metadata"]


# Archives and file links used to download Clotho
_CLOTHO_LINKS: Dict[
    ClothoVersion, Dict[ClothoSubset, Dict[_ClothoLinkType, LinkInfoHash]]
] = {
    "v1": {
        "dev": {
            "audio_archive": {
                "fname": "clotho_audio_development.7z",
                "url": "https://zenodo.org/record/3490684/files/clotho_audio_development.7z?download=1",
                "hash_value": "e3ce88561b317cc3825e8c861cae1ec6",
            },
            "captions": {
                "fname": "clotho_captions_development.csv",
                "url": "https://zenodo.org/record/3490684/files/clotho_captions_development.csv?download=1",
                "hash_value": "dd568352389f413d832add5cf604529f",
            },
            "metadata": {
                "fname": "clotho_metadata_development.csv",
                "url": "https://zenodo.org/record/3490684/files/clotho_metadata_development.csv?download=1",
                "hash_value": "582c18ee47cebdbe33dce1feeab53a56",
            },
        },
        "eval": {
            "audio_archive": {
                "fname": "clotho_audio_evaluation.7z",
                "url": "https://zenodo.org/record/3490684/files/clotho_audio_evaluation.7z?download=1",
                "hash_value": "4569624ccadf96223f19cb59fe4f849f",
            },
            "captions": {
                "fname": "clotho_captions_evaluation.csv",
                "url": "https://zenodo.org/record/3490684/files/clotho_captions_evaluation.csv?download=1",
                "hash_value": "1b16b9e57cf7bdb7f13a13802aeb57e2",
            },
            "metadata": {
                "fname": "clotho_metadata_evaluation.csv",
                "url": "https://zenodo.org/record/3490684/files/clotho_metadata_evaluation.csv?download=1",
                "hash_value": "13946f054d4e1bf48079813aac61bf77",
            },
        },
        "dcase_aac_test": {
            "audio_archive": {
                "fname": "clotho_audio_test.7z",
                "url": "https://zenodo.org/record/3865658/files/clotho_audio_test.7z?download=1",
                "hash_value": "9b3fe72560a621641ff4351ba1154349",
            },
            "metadata": {
                "fname": "clotho_metadata_test.csv",
                "url": "https://zenodo.org/record/3865658/files/clotho_metadata_test.csv?download=1",
                "hash_value": "52f8ad01c229a310a0ff8043df480e21",
            },
        },
    },
    "v2": {
        "dev": {
            "audio_archive": {
                "fname": "clotho_audio_development.7z",
                "url": "https://zenodo.org/record/4743815/files/clotho_audio_development.7z?download=1",
                "hash_value": "eda144a5e05a60e6d2e37a65fc4720a9",
            },
            "captions": {
                "fname": "clotho_captions_development.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_captions_development.csv?download=1",
                "hash_value": "800633304e73d3daed364a2ba6069827",
            },
            "metadata": {
                "fname": "clotho_metadata_development.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_metadata_development.csv?download=1",
                "hash_value": "5fdc51b4c4f3468ff7d251ea563588c9",
            },
        },
        "val": {
            "audio_archive": {
                "fname": "clotho_audio_validation.7z",
                "url": "https://zenodo.org/record/4743815/files/clotho_audio_validation.7z?download=1",
                "hash_value": "0475bfa5793e80f748d32525018ebada",
            },
            "captions": {
                "fname": "clotho_captions_validation.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_captions_validation.csv?download=1",
                "hash_value": "3109c353138a089c7ba724f27d71595d",
            },
            "metadata": {
                "fname": "clotho_metadata_validation.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_metadata_validation.csv?download=1",
                "hash_value": "f69cfacebcd47c4d8d30d968f9865475",
            },
        },
        "eval": {
            "audio_archive": {
                "fname": "clotho_audio_evaluation.7z",
                "url": "https://zenodo.org/record/4743815/files/clotho_audio_evaluation.7z?download=1",
                "hash_value": "4569624ccadf96223f19cb59fe4f849f",
            },
            "captions": {
                "fname": "clotho_captions_evaluation.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_captions_evaluation.csv?download=1",
                "hash_value": "1b16b9e57cf7bdb7f13a13802aeb57e2",
            },
            "metadata": {
                "fname": "clotho_metadata_evaluation.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_metadata_evaluation.csv?download=1",
                "hash_value": "13946f054d4e1bf48079813aac61bf77",
            },
        },
        "dcase_aac_test": {
            "audio_archive": {
                "fname": "clotho_audio_test.7z",
                "url": "https://zenodo.org/record/3865658/files/clotho_audio_test.7z?download=1",
                "hash_value": "9b3fe72560a621641ff4351ba1154349",
            },
            "metadata": {
                "fname": "clotho_metadata_test.csv",
                "url": "https://zenodo.org/record/3865658/files/clotho_metadata_test.csv?download=1",
                "hash_value": "52f8ad01c229a310a0ff8043df480e21",
            },
        },
    },
    "v2.1": {
        "dev": {
            "audio_archive": {
                "fname": "clotho_audio_development.7z",
                "url": "https://zenodo.org/record/4783391/files/clotho_audio_development.7z?download=1",
                "hash_value": "c8b05bc7acdb13895bb3c6a29608667e",
            },
            "captions": {
                "fname": "clotho_captions_development.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_captions_development.csv?download=1",
                "hash_value": "d4090b39ce9f2491908eebf4d5b09bae",
            },
            "metadata": {
                "fname": "clotho_metadata_development.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_metadata_development.csv?download=1",
                "hash_value": "170d20935ecfdf161ce1bb154118cda5",
            },
        },
        "val": {
            "audio_archive": {
                "fname": "clotho_audio_validation.7z",
                "url": "https://zenodo.org/record/4783391/files/clotho_audio_validation.7z?download=1",
                "hash_value": "7dba730be08bada48bd15dc4e668df59",
            },
            "captions": {
                "fname": "clotho_captions_validation.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_captions_validation.csv?download=1",
                "hash_value": "5879e023032b22a2c930aaa0528bead4",
            },
            "metadata": {
                "fname": "clotho_metadata_validation.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_metadata_validation.csv?download=1",
                "hash_value": "2e010427c56b1ce6008b0f03f41048ce",
            },
        },
        "eval": {
            "audio_archive": {
                "fname": "clotho_audio_evaluation.7z",
                "url": "https://zenodo.org/record/4783391/files/clotho_audio_evaluation.7z?download=1",
                "hash_value": "4569624ccadf96223f19cb59fe4f849f",
            },
            "captions": {
                "fname": "clotho_captions_evaluation.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_captions_evaluation.csv?download=1",
                "hash_value": "1b16b9e57cf7bdb7f13a13802aeb57e2",
            },
            "metadata": {
                "fname": "clotho_metadata_evaluation.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_metadata_evaluation.csv?download=1",
                "hash_value": "13946f054d4e1bf48079813aac61bf77",
            },
        },
        "dcase_aac_test": {
            "audio_archive": {
                "fname": "clotho_audio_test.7z",
                "url": "https://zenodo.org/record/3865658/files/clotho_audio_test.7z?download=1",
                "hash_value": "9b3fe72560a621641ff4351ba1154349",
            },
            "metadata": {
                "fname": "clotho_metadata_test.csv",
                "url": "https://zenodo.org/record/3865658/files/clotho_metadata_test.csv?download=1",
                "hash_value": "52f8ad01c229a310a0ff8043df480e21",
            },
        },
        "dcase_aac_analysis": {
            "audio_archive": {
                "fname": "clotho_analysis_2022.zip",
                "url": "https://zenodo.org/record/6610709/files/clotho_analysis_2022.zip?download=1",
                "hash_value": "7e8fa4762cc3a7c5546606680b958d08",
            },
        },
        "dcase_t2a_audio": {
            "audio_archive": {
                "fname": "retrieval_audio.7z",
                "url": "https://zenodo.org/record/6590983/files/retrieval_audio.7z?download=1",
                "hash_value": "24102395fd757c462421a483fba5c407",
            },
            "metadata": {
                "fname": "retrieval_audio_metadata.csv",
                "url": "https://zenodo.org/record/6590983/files/retrieval_audio_metadata.csv?download=1",
                "hash_value": "1301db07acbf1e4fabc467eb54e0d353",
            },
        },
        "dcase_t2a_captions": {
            "captions": {
                "fname": "retrieval_captions.csv",
                "url": "https://zenodo.org/record/6590983/files/retrieval_captions.csv?download=1",
                "hash_value": "f9e810118be00c64ea8cd7557816d4fe",
            },
        },
    },
}

# Captions column names in CSV files
_CAPTIONS_KEYS = (
    "caption_1",
    "caption_2",
    "caption_3",
    "caption_4",
    "caption_5",
)

# Metadata column names in CSV files
_METADATA_KEYS = (
    "keywords",
    "sound_id",
    "sound_link",
    "start_end_samples",
    "manufacturer",
    "license",
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module containing the datasets classes source code for :class:`~aac_datasets.datasets.audiocaps.AudioCaps`, :class:`~aac_datasets.datasets.clotho.Clotho`, :class:`~aac_datasets.datasets.macs.MACS` and :class:`~aac_datasets.datasets.wavcaps.WavCaps`.
"""

from .audiocaps import AudioCaps
from .clotho import Clotho
from .macs import MACS
from .wavcaps import WavCaps



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Literal, Tuple, get_args, overload

from .audiocaps import AudioCaps
from .base import AACDataset
from .clotho import Clotho
from .macs import MACS
from .wavcaps import WavCaps

DatasetName = Literal["AudioCaps", "Clotho", "MACS", "WavCaps"]


def list_datasets_names() -> Tuple[DatasetName, ...]:
    return get_args(DatasetName)


@overload
def load_dataset(name: Literal["AudioCaps"], *args, **kwargs) -> AudioCaps: ...


@overload
def load_dataset(name: Literal["Clotho"], *args, **kwargs) -> Clotho: ...


@overload
def load_dataset(name: Literal["MACS"], *args, **kwargs) -> MACS: ...


@overload
def load_dataset(name: Literal["WavCaps"], *args, **kwargs) -> WavCaps: ...


def load_dataset(name: DatasetName, *args, **kwargs) -> AACDataset:
    if name == "AudioCaps":
        return AudioCaps(*args, **kwargs)
    elif name == "Clotho":
        return Clotho(*args, **kwargs)
    elif name == "MACS":
        return MACS(*args, **kwargs)
    elif name == "WavCaps":
        return WavCaps(*args, **kwargs)
    else:
        msg = f"Invalid argument {name=}. (expected one of {get_args(DatasetName)})"
        raise ValueError(msg)
